/*************************************************************************
 * ateamp.c - Main function for the parallel Asynchronous Team
 * for the point-to-point connection problem
 * 
 * (C) 1996-1998 Carlos Alberto de S. Oliveira
 ************************************************************************/

#include <pvm3.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "ateams.h"
#include "sinc.h"

/* store solutions beeing processed in parallel */
static pending_sol_t pending_sol[MAX_PAR_ASYNC];
/* pending solutions (from 0 to MAX_PAR_ASYNC)  */
static int pending_write = 0;
/* number of messages sent waiting for an ack   */
int expected;
/* name of the parameter file                   */
static char param_file[256];
/* name of lock file                            */
static char lockf_name[BN_SIZE];
/* the number of current pulse of execution     */
static int g_current_pulse;
/* number of work_done messages received        */
int n_wd_messages = 0;
/* lock file                                    */
FILE *g_lock;
/* number of neigboors                          */
int n_neighbors;
/* process identifiers                          */
int p_tid, my_tid, my_num;
/* vector of neighbors to this node             */
int *neig, *delta;
/* true if we already processed work done msg   */
int work_done_sent = 0;
int x_torus;
int paths_received=0;
int nx_paths = 0, ny_paths=0;
paths_description *paths_descr;

/* clean up the program and prepares to finish */
void finish_execution()
{
  pvm_initsend(PvmDataDefault);
  pvm_send(p_tid,termination);
  Delete(neig);
  Delete(delta);
  fprintf(g_of, "End of processing. Total time: %ds\n",
	  (int)difftime(time(NULL), gTimeIni));
  fclose(g_lock);
  if (my_num == 0)
    remove(lockf_name);
  else
    remove(g_out_file);
  fclose(g_of);
  pvm_exit();
}

int last_char_pos(char *str, char c)
{
  int i, resp;
  for (i=0; *str; i++, str++)
    if (*str == c)
      resp = i;
  return resp;
}

void create_output_file(char *param_file)
{
  int len, slash_pos, point_pos;
  char rel_path[BN_SIZE];
  char base_name[BN_SIZE];
  
  /* find relative path */
  len = strlen(param_file);
  slash_pos = last_char_pos(param_file, '/');
  strncpy(rel_path, param_file, slash_pos+1);
  rel_path[slash_pos+1] = '\x0';
  
  /* create a lock file */
  sprintf(lockf_name, "%sppclock", rel_path);
  g_lock = fopen(lockf_name, "w+");
  if (!g_lock)
    StdErro("P: Can't open lock file.\n");
  
  /* create an output file */
  point_pos = last_char_pos(param_file, '.');
  if (point_pos < slash_pos)
    point_pos = strlen(param_file);
  strncpy(base_name, param_file+slash_pos+1, point_pos-slash_pos-1);
  base_name[point_pos-slash_pos-1] = '\x0';
  sprintf(g_out_file,  "%s%s-p%x.txt",
	  rel_path, base_name, my_tid);
  g_of = fopen(g_out_file, "w");
  if (!g_of)
    StdErro("P: Can't open output file\n");
}

void print_parallel_results()
{
  result_t *r;
  fprintf(g_of, "n.iter\tBestSol\tHeurist.\tProcessor\n");
  for (r = result->next; r; r = r->next)
    fprintf(g_of, "%d\t%d\t%s\t%d\n", r->iter, r->best,
	    heuristic[r->heurist].h_name, r->processor);
}

/* defines the number of elements in the reach for asking
   parameters. This can be changed to an appropriate function. */
int reach_size()
{
  return randomico(n_neighbors+1);
}

/* insert a parameter received from other processors in
   `pending_sol[id]'. The values returned must be ordered in order
   to use the standard write selection procedure */
void insert_new_parameter(int id, int position, int value, int n_orig)
{
  int i, n, *reach, *values, *pos;
  n = pending_sol[id].size_reach;
  reach = pending_sol[id].reach;
  pos = pending_sol[id].pos;
  values = pending_sol[id].value;
  for (i=0; i<n; i++){
    if (reach[i] == n_orig)
      break;
  }
  if (i==n){
    fprintf(g_of, "Error receiving parameter. n=%d, id=%d,"
	    "norig = %d\n", n, id, n_orig);
    finish_execution();
    exit(1);
  }
  while (i>0 && value < values[i-1]) {
    pos[i] = pos[i-1];
    values[i] = values[i-1];
    reach[i] = reach[i-1];
    i--;
  }
  while (i<n && value > values[i+1]) {
    pos[i] = pos[i+1];
    values[i] = values[i+1];
    reach[i] = reach[i+1];
    i++;
  }
  reach[i] = n_orig;
  values[i] = value;
  pos[i] = position;
  pending_sol[id].n_received++;
}

/* send a solution to processor `dest' to be put in position `pos' */
void send_solution(Solution *sol, int p, int dest, int iter, int pos)
{
  int i, value;
  listvert *l;
  begin_sendmsg(dest, receive_sol);
  pvm_pkint(&iter, 1, 1);
  pvm_pkint(&pos, 1, 1);
  for (i=0; i<p; i++){
    /* send destination and path */
    pvm_pkint(&(sol[i].dest), 1, 1);
    for (l=sol[i].path; l; l=l->prox)
      pvm_pkint(&(l->v), 1, 1);
    value = -1;
    pvm_pkint(&value, 1, 1);
  }
  end_sendmsg;
  pending_write++;
}

/* receives a solution sent by a neighbor and insert it
   if its acceptable */
void receive_solution(int source)
{
  Solution *sol;
  listvert *l;
  int p = gInst->p, i, v, changed, remote_iter, pos;

  sol = NewVec(Solution, p);
  if (!sol)
    StdErro("error: could not allocate memory to receive "
	     "new solution.\n");

  /* read iteration in which the solution was created */
  pvm_upkint(&remote_iter, 1, 1);
  pvm_upkint(&pos, 1, 1);
  /* read the solution */
  for (i=0; i<p; i++){
    pvm_upkint(&(sol[i].dest), 1, 1);
    sol[i].path = 0;
    while (pvm_upkint(&v, 1, 1), v != -1){
      if (sol[i].path == 0){
	sol[i].path = new_list_vert();
	l = sol[i].path;
      } else {
	l->prox = new_list_vert();
	l = l->prox;
      }
      l->v = v;
      l->prox = 0;
    }
  }
  /* verify if solution is correct */
  if (!validate_solution(sol)) {
    fprintf(g_of, "solution received is invalid!\n");
    fflush(g_of);
  }
  /* insert the new solution in local memory */
  insert_new_solution(sol, gInst->G, gInst->p, pos);
  gNumIter = max(gNumIter, remote_iter);
  send_msg(source, end_write);
  changed = update_best_solution();
}

/* execute an iteration of the A-Team */
void do_parallel_computation(Graph *G, vert s[], vert d[], int p)
{
  Solution *sol;
  int       pos, n_neig, rnd;
  int       i, s_index;

  gNumIter++;
  g_heurist_chosen = select_random_heuristic();
  heuristic[g_heurist_chosen].n_used++;
  /* run the selected heuristic */
  sol = (*heuristic[g_heurist_chosen].func)(G, s, d, p);
  /* decides where to put the solution */
  n_neig = reach_size();
  if (n_neig != 0) {
    /* find an empty slot in pending_sol */
    for (i=0; i<MAX_PAR_ASYNC; i++)
      if (!pending_sol[i].s){
	s_index = i;
	break;
      }
    if (i==MAX_PAR_ASYNC){
      fprintf(g_of, "Error: more pending solutions than expected.\n");
      fflush(g_of);
      finish_execution();
      exit(1);
    }
    memset(pending_sol[s_index].reach, 0, sizeof(int)*MAX_NUM_NEIGHBORS);
    /* request the param from all my neighbors */
    for (i=0, pos=-1; i<n_neig; i++){
      rnd = rand() % (n_neighbors - (n_neig-i-1) - pos - 1);
      pos = 1 + pos + rnd;
      pending_sol[s_index].reach[i] = neig[pos];
      begin_sendmsg(neig[pos], ask_param);
      pvm_pkint(&s_index, 1, 1); /* identifier for solution sent */
      end_sendmsg;
    }
    pending_sol[s_index].reach[i] = my_num;
    pending_sol[s_index].s = copy_solution(sol, p);
    pending_sol[s_index].iter = gNumIter;
    pending_sol[s_index].n_received = 0;
    pending_sol[s_index].size_reach = n_neig+1;
  } else {
    insert_new_solution(sol, G, p, -1);
    update_best_solution();
  }
}

/* send the response to processor `source'*/
void send_parameter(int source)
{
  int pos, value, id;
  pvm_upkint(&id, 1, 1);
  pos = get_delete_position();
  value = solucoes[pos].custo;
  begin_sendmsg(source, resp_param);
  pvm_pkint(&id, 1, 1);
  pvm_pkint(&pos, 1, 1);
  pvm_pkint(&value, 1, 1);
  end_sendmsg;
}

/* process all parameters received in response to a request */
void process_parameters (int id)
{
  int p=gInst->p, pos, local_pos;
  int n = pending_sol[id].size_reach;

  local_pos = get_delete_position();
  insert_new_parameter(id, local_pos, solucoes[local_pos].custo, my_num);
  /* find where the solution will be inserted */
  pos = linear_rand(n)-1;
  if (pending_sol[id].reach[pos] == my_num){
    insert_new_solution(pending_sol[id].s, gInst->G, p, pos);
  } else {
    /* write solution in other processor */
    send_solution(pending_sol[id].s, p, pending_sol[id].reach[pos],
		  pending_sol[id].iter, pending_sol[id].pos[pos]);
    DeleteSolution(pending_sol[id].s, p);
  }
  pending_sol[id].s = 0;
}

/* process an incomming parameter from processor `n_orig' */
void receive_parameter(int n_orig)
{
  int position, value, id;
  pvm_upkint(&id,    1, 1);
  pvm_upkint(&position, 1, 1);
  pvm_upkint(&value, 1, 1);
  insert_new_parameter(id, position, value, n_orig);
  if (pending_sol[id].n_received == pending_sol[id].size_reach-1)
    process_parameters(id);
}

/* process the reception of message `work_done' */
void receive_work_done_message()
{
  result_t *r, *s, *t;
  int iter, best, heurist, processor;
  n_wd_messages++;

  if (n_wd_messages == 1) {
    for (r=result->next; r; r=r->next)
      r->processor = my_num;
  }
  while (pvm_upkint(&iter, 1, 1), iter != -1 ) {
    pvm_upkint(&best, 1, 1);
    pvm_upkint(&heurist, 1, 1);
    pvm_upkint(&processor, 1, 1);
    r = result;
    /* find the right position to insert           */
    while (r->next && r->next->iter < iter)
      r = r->next;
    /* insert only if this makes an improvement    */
    if (r==result || (r != result && r->best >  best)){
      /* delete solutions that are worse than this */
      while (r->next && r->next->best >= best) {
	s = r->next->next;
	Delete(r->next);
	r->next = s;
      }
      t = New(result_t);
      if (!t)
	return;
      t->iter = iter;
      t->best = best;
      t->heurist = heurist;
      t->processor = processor;
      t->next = r->next;
      r->next = t;  
    }
  }
}

/* verify the `work_done' messages already received. 
   returns 0 if the program continues, else returns 1 */
int process_work_done_message()
{
  int res = 0, i;
  result_t *r;
  if (my_num == 0) {
    if (n_wd_messages == n_neighbors) {
      for (i=0; i<n_neighbors; i++)
	send_msg(neig[i], terminate);
      print_parallel_results();
      res = 1;
      work_done_sent = 1;
    }
  } else {
    begin_sendmsg(0, work_done);
    /* send all results */
    for (r = result->next; r; r=r->next) {
      pvm_pkint(&(r->iter), 1, 1);
      pvm_pkint(&(r->best), 1, 1);
      pvm_pkint(&(r->heurist), 1, 1);
      pvm_pkint(&my_num, 1, 1);
    }
    i = -1;
    pvm_pkint(&i, 1, 1);
    end_sendmsg;
    work_done_sent = 1;
  }
  return res;
}

void calc_shortest_paths()
{
  int i, c, n, beg, end, x, dest, v=1;
  Graph *G = gInst->G;
  n = G_nVerts(G);
  /* number of vertices I'll process */
  c = n/(n_neighbors+1);
  beg = (c*my_num)+1;
  end = c*(my_num+1);
  if (my_num == n_neighbors)
    end = n;
  for (i=beg; i<=end; i++){
    all_shortest_paths_from_vertex(G, i, A[i], Pi[i]);
  }
  /* number of lines in the torus configuration */
  x = (int) sqrt(n_neighbors+1);
  while ((n_neighbors+1)%x != 0)
    x--;
  if (x == 1){
    fprintf(g_of, "The number of processors can not be prime.\n");
    finish_execution();
    exit(1);
  }
  x_torus = x;
  paths_descr = NewVec(paths_description, x);
  paths_descr[0].beg = beg;
  paths_descr[0].end = end;
  if ((my_num+1)%x == 0)
    dest = (my_num/x)*x;
  else
    dest = my_num+1;
  begin_sendmsg(dest, send_paths1);
  pvm_pkint(&beg, 1, 1);
  pvm_pkint(&end, 1, 1);
  pvm_pkint(&v, 1, 1);
  for (i=beg; i<=end; i++){
    pvm_pkint(A[i], n+1, 1);
    pvm_pkint(Pi[i], n+1, 1);
  }
  end_sendmsg;
  nx_paths = 1;
}

/* send min paths in the other direction */
void process_ytorus()
{
  int i, j, v, x=x_torus, dest, beg, end, n=G_nVerts(gInst->G);
  /* send in the other direction */
  dest = my_num + x;
  dest = dest > n_neighbors ? dest % x : dest;
  begin_sendmsg(dest, send_paths2);
  v = 1;
  pvm_pkint(&v, 1, 1);
  for (i=0; i<x; i++){
    beg = paths_descr[i].beg;
    end = paths_descr[i].end;
    pvm_pkint(&beg, 1, 1);
    pvm_pkint(&end, 1, 1);
    for (j=beg; j<=end; j++){
      pvm_pkint(A[j], n+1, 1);
      pvm_pkint(Pi[j], n+1, 1);
    }
  }
  end_sendmsg;
  ny_paths = 1;
}

void receive_path1()
{
  int v, beg, end, dest, i, x=x_torus, n=G_nVerts(gInst->G);
  pvm_upkint(&beg, 1, 1);
  pvm_upkint(&end, 1, 1);
  pvm_upkint(&v, 1, 1);
  for (i=beg; i<=end; i++) {
    pvm_upkint(A[i], n+1, 1);
    pvm_upkint(Pi[i], n+1, 1);
  }
  paths_descr[nx_paths].beg = beg;
  paths_descr[nx_paths].end = end;
  v++;
  if (v < x) {
    if ((my_num+1)%x == 0 || my_num == n_neighbors){
      dest = (my_num/x)*x;
    } else
      dest = my_num+1;
    begin_sendmsg(dest, send_paths1);
    pvm_pkint(&beg, 1, 1);
    pvm_pkint(&end, 1, 1);
    pvm_pkint(&v, 1, 1);
    for (i=beg; i<=end; i++){
      pvm_pkint(A[i], n+1, 1);
      pvm_pkint(Pi[i], n+1, 1);
    }
    end_sendmsg;
  }
  nx_paths++;
  if (nx_paths == x)
    process_ytorus();
}

void receive_path2()
{
  int v, beg, end, dest, i, j, x=x_torus, n=G_nVerts(gInst->G);
  dest = my_num+x;
  dest = dest > n_neighbors ? dest%x : dest;
  pvm_upkint(&v, 1, 1);
  for (i=0; i<x; i++) {
    pvm_upkint(&beg, 1, 1);
    pvm_upkint(&end, 1, 1);
    for (j=beg; j<=end; j++) {
      pvm_upkint(A[j], n+1, 1);
      pvm_upkint(Pi[j], n+1, 1);
    }
    paths_descr[i].beg = beg;
    paths_descr[i].end = end;
  }
  v++;
  if (v < (n_neighbors+1)/x){
    begin_sendmsg(dest, send_paths2);
    pvm_pkint(&v, 1, 1);
    for (i=0; i<x; i++){
      beg = paths_descr[i].beg;
      end = paths_descr[i].end;
      pvm_pkint(&beg, 1, 1);
      pvm_pkint(&end, 1, 1);
      for (j=beg; j<=end; j++){
	pvm_pkint(A[j], n+1, 1);
	pvm_pkint(Pi[j], n+1, 1);
      }
    }
    end_sendmsg;
  }
  ny_paths++;
  if (ny_paths == (n_neighbors+1)/x){
    paths_received = 1;
    do_parallel_computation(gInst->G, gInst->source,
				gInst->dest, gInst->p);
  }
}

void do_initial_computation()
{
  /* Read the instance file */
  gInst = read_instance (param_file);
  if (!gInst)
    StdErro("Terminando por problemas na leitura de arquivo\n");
  strcpy(gInst->instName, param_file);
  result = last_res = New(result_t);
  result->next = 0;
  /* Initialize data structures */
   gNumSolutions = (gNumSolutions =  G_nVerts(gInst->G)) % 2 ?
     gNumSolutions : gNumSolutions + 1;
  gNumSolutions = (gNumSolutions)/(n_neighbors+1);
 
  init_solutions();
  init_shared_memory(gInst->G, gInst->source, gInst->dest, gInst->p);
  memset(pending_sol, 0, sizeof(pending_sol_t)*MAX_PAR_ASYNC);
  gBestSolValue  = solucoes[best_solution_pos].custo;
  gWorstSolValue = solucoes[worst_solution_pos].custo;
  calc_shortest_paths();
}

int main()
{
  int end, n, rand_value;
  int norig, pulse_origin;

  gTimeIni = time(NULL);
  /* Inicialization */
  p_tid = pvm_parent();
  my_tid = pvm_mytid();
  gVerbose = 0;
  end      = 0;

  while (!end) {
    int        bufid, len, who, tipo;
    msglabel_t type;

    /* Getting a Message */
    bufid = pvm_recv(p_tid, -1);
    pvm_bufinfo(bufid, &len, (int *) &type, &who);

    switch (type) {
    case startup:
      pvm_upkint(&n, 1, 1);
      n_neighbors = n;
      neig  = NewVec(int, n);
      delta = NewVec(int, n);
      pvm_upkint(neig, n, 1);
      pvm_upkint(delta, n, 1);
      pvm_upkstr(param_file);
      pvm_upkint(&my_num, 1, 1);
      pvm_upkint(&rand_value, 1, 1);
      /* initialize the random generator */
      srand(rand_value);

      create_output_file(param_file);
      /* write information */
#if 0      
      {
	int i;      
	fprintf(g_of,"Node %d, %d neighbors, father %x\n"
		"neighbors ", my_num, n, p_tid);
	for(i=0;i<n;i++)
	  fprintf(g_of,"%d, ", neig[i]); 
	fprintf(g_of,"\n"); fflush(g_of);
      }
#endif      
      do_initial_computation();
      /*do_parallel_computation(gInst->G, gInst->source,
			      gInst->dest, gInst->p);*/
      pvm_send(p_tid, safe); 
      break;
      
    case safe:
      pvm_upkint(&g_current_pulse,1,1);
      pvm_upkint(&norig,1,1);
      
      while(norig > -1) {
	pvm_upkint(&pulse_origin,1,1);
	pvm_upkint(&tipo,1,1);

	switch(tipo) {
	case send_paths1:
	  receive_path1();
	  break;
	case send_paths2:
	  receive_path2();
	  break;
	case ask_param:
	  send_parameter(neig[norig]); 
	  break;
	case resp_param:
	  receive_parameter(neig[norig]);
	  break;
	case receive_sol:
	  receive_solution(neig[norig]);
	  break;
	case end_write:
	  pending_write--;
	  break;
	case work_done:
	  if (my_num == 0)
	    receive_work_done_message();
	  break;
	case terminate:
	  end = 1;
	  break;
	default:
	  fprintf(g_of, "Invalid message %d.\n", tipo);
	  fflush(g_of);
	}
	pvm_upkint(&norig,1,1); 
      }
      
      if (gNumIter >= g_max_num_iter){
	if (pending_write == 0 && !work_done_sent){
	  end = process_work_done_message();
	}
      } else if (paths_received)
	do_parallel_computation(gInst->G, gInst->source,
				gInst->dest, gInst->p);
      pvm_send(p_tid, safe);
      break;

    case ack:
      expected--;
    case neigmsg:
    case compmsg:
    case termination:
    }
  }
  finish_execution();
  return 0;
}
