/*----------------------------------------------------------------
 *  ATeams
 *  Program to solve the Point-to-point connection problem
 *
 *  (C)Carlos A.S. Oliveira 1996-2001
 *----------------------------------------------------------------*/

/* ateams.h: General declarations and data structures */

#ifndef _ATEAMS_H
#define _ATEAMS_H

#include <time.h>
#include <limits.h>

#ifdef HAVE_MINGW32
#define _WINDOWS_
#endif

#ifdef _WINDOWS_
#include <windows.h>
#endif

#include "common.h"

#define PARALLEL
#include "graph.h"

#define INF                  INT_MAX /* infinity */
#define MAX_NUM_HEURISTICS   64

#define fatal_error(msg)    \
do {                    \
  fprintf(g_of, msg);   \
  if (g_of != stdout)   \
    fclose(g_of);       \
  exit (1);             \
} while (0)

/* prints a message made of multiple lines, ended by a empty string */
#define mlprint(c)             \
do {                           \
	int i = 0;             \
	while (strlen (c[i]))  \
		puts (c[i++]); \
} while (0)

#define begin_sendmsg(dest, type)                            \
do {                                                         \
  int d = (dest), t = (type);                                \
  pvm_initsend(PvmDataDefault);  /* init send            */  \
  pvm_pkint(&d, 1, 1);           /* destination neighbor */  \
  pvm_pkint(&t, 1, 1);           /* type of message      */  \
} while (0)

#define end_sendmsg          \
do {                         \
  pvm_send(p_tid, compmsg);  \
  expected++;                \
} while (0)

#define send_msg(dest, type)                                 \
do {                                                         \
  int d = (dest), t = (type);                                \
  pvm_initsend(PvmDataDefault);  /* init send            */  \
  pvm_pkint(&d, 1, 1);           /* destination neighbor */  \
  pvm_pkint(&t, 1, 1);           /* type of message      */  \
  pvm_send(p_tid, compmsg);                                  \
  expected++;                                                \
} while (0)

#define best_solution_pos    0
#define worst_solution_pos  (gNumSolutions-1)

#define MAX_PAR_ASYNC     32
#define MAX_NUM_NEIGHBORS 32
#define BN_SIZE           128

#ifdef PARALLEL
/*
 * types of messages used in the parallel
 * implementation
 */
typedef enum {
  send_paths1,
  send_paths2,
  ask_param,
  resp_param,
  receive_sol,
  end_write,
  work_done,
  terminate
} t_message;
#endif


/* the PPC instance */
typedef struct {
  char instName[200];	   /* name of the instance file */
  char out_file_name[200]; /* name of the output file   */
  int p;		   /* number of sources and destinations */
  Graph *G;		   /* The graph itself */
  vert *source;		   /* the source vertex set */
  vert *dest;		   /* Destination vertex set */
  int hasBound;		   /* Indicates if there are upper and lower
			      bounds to this instance */
  int Lower;		   /* Lower bound of the instance */
  int Upper;		   /* Upper bound of the instance */
  FILE *of;                /* Output file                 */
} ppc_instance;

/* used to build vertices */
typedef struct taglistvert{
  vert v;			/* vertex */
  struct taglistvert *prox;	/* link */
} listvert;

/* a cell of a solution array */
typedef struct {
  vert dest;			/* destination vertex */
  listvert *path;		/* path to the destination */
} Solution;

/* a complete solution */
typedef struct {
  int custo;			/* Cost of the solution */
  Solution *sol;		/* Link to the solution */
} raiz;

typedef struct _result_t {
  int iter;
  int best;
  int heurist;
  int processor;
  struct _result_t *next;
} result_t;

/* an arest */
typedef struct {
  vert a;
  vert b;
} arest;

/* an partial memory used to store a set of arests */
typedef struct {
  arest *ar;       /* an array of arests */
  int n;           /* number of arests stored */
} partMem;

typedef struct {
  char *h_name;      /* name of heuristic                              */
  int n_used;        /* times it was used                              */
  Solution *(*func)(Graph*,vert[],vert[],int); /* pointer to heuristic */
} t_heuristic;

typedef struct {
  Solution *s;
  int iter;
  int n_received;
  int pos[MAX_NUM_NEIGHBORS];
  int value[MAX_NUM_NEIGHBORS];
  int size_reach;
  int reach[MAX_NUM_NEIGHBORS];
} pending_sol_t;

typedef struct {
  int beg;
  int end;
  int v;
} paths_description;

typedef struct {
  int *a;
  int n;
} cond_in_set_type;


#define VERBOSE(str) \
  if (gVerbose) fprintf(g_of, str), fflush(g_of)

#ifdef _WINDOWS_

#define protect_thread    WaitForSingleObject(gMutex, INFINITE)
#define unprotect_thread  if (!ReleaseMutex(gMutex)) ErrorMsg("Mutex erro");
#else

#define protect_thread
#define unprotect_thread
     
#endif     

#include "rf.h"

/*
 * Functions prototypes
 */

listvert *new_list_vert (void);
void delete_list_vert (listvert * l);
listvert *shortest_path (Graph * G, vert v, vert dest);
listvert *random_path(Graph * G, vert v, vert dest);
listvert *random_path_sdfree(Graph *G, vert *v, vert *select_dest,
                             vert *source, vert *dest, int n_sources);



void delete_path(listvert *h);
void print_list(listvert * s);
void print_solution(FILE * f, Solution * sol, int p);
Graph *mst (void);
int get_delete_position(void);
void DeleteSolution (Solution * s, int p);

/*
 * Heuristics implemented in this A-Team
 * for each heuristic the parameters are:
 *  G - Graph
 *  s - vector of source nodes
 *  d - vector of destination nodes
 *  p - number of source-destination pairs.
 */
Solution *Heuristic1 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic2 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic3 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic4 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic5 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic6 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic7 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic8 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic9 (Graph *G, vert s[], vert d[], int p);
Solution *Heuristic10(Graph *G, vert s[], vert d[], int p);
Solution *Heuristic11(Graph *G, vert s[], vert d[], int p);
Solution *Heuristic12(Graph *G, vert s[], vert d[], int p);

Solution *constructor2(Graph*, vert[], vert[], int);
Solution *permute_paths (Solution * sol, Graph * G, int s[], int p);

int arests_in_solution(Solution * sol, arest a[], int p);
void init_solutions (void);
int IsEqual (Solution * a, Solution * b, int p);
Solution *copy_solution (Solution * s, int p);
int eval_solution (Solution * s, Graph * G, int p);
ppc_instance *read_instance(char *sFileName);
int validate_solution (Solution * sol);
void insert_solution(Solution * sol, int pos, int custo);
int delete_loop (listvert * path, Graph * G);
void DeleteSolutions (int p);
void init_ateam(Graph * G, vert s[], vert d[], int p);
void run_ateam(Graph *G, vert s[], vert d[], int p);
void ProcessCommandLineArgs (int n, char *arg[]);
void InitHeurist();
void DeleteHeurist();
void DeleteListvertModule();
void InitListvertModule();
void DeletePaths();
int InitPaths();
int can_stop(int best);
int SelectRandomHeuristic();
void CheckInstanceBounds(FILE *f, int better);
void ShowNumbHeuristicExecution(FILE *f);
void insert_new_solution(Solution *sol, Graph *G, int p, int pos);
void do_some_computation(Graph *G, vert s[], vert d[], int p);
int update_best_solution(void);
void show_ateam_results(int changed);
void init_shared_memory(Graph *G, vert s[], vert d[], int p);
int shortest_path_dist(Graph * G, vert source, vert d);
int mst_bound(int argc, char **argv);
void calc_arest_usage(Graph *G, Graph *ng, Solution *sol, int p);
int get_num_heuristics(void);
int select_random_heuristic(void);
void all_shortest_paths_from_vertex(Graph*, vert,int *, int *);
void all_shortest_paths(Graph *G);
int linear_rand(int);
int remove_no_sdleaves(Graph *G, int s[], int d[], int p);
Solution *rf_create_solution_from_graph(restricted_forest_tad *rf,
					Graph *G);
Graph *restricted_forest(restricted_forest_tad *rf, float* lb);
int cond_is_in_set(int v, cond_in_set_type *t);

/*-----------------------------
 *  External variables
 *----------------------------*/

/* flags for program execution */
extern int gReadMem, gInitMem, gVerbose;
/* file to initialize the memory */
extern char gReadArq[];
/* name of the instance definition file */
extern char g_out_file[];
/* Default output file */
extern FILE *g_of;

/* some parameters */ 
extern int gMaxTime, gOptimum;
extern int gNumSolutions, gNumRepeat;
/* inicial time for A-Team */
time_t gTimeIni;

/* last heuristic used */
extern int g_heurist_chosen;

/* list of heuristics used in this execution and they number*/
extern int gHeuristList[], gNumHeurist;

/* The global instance of the problem */
extern ppc_instance *gInst;
/* shared memory */
extern raiz *solucoes;
/* shared partial memory */
extern partMem PartialMem;
/* value of best and worst solution */
extern int gBestSolValue, gWorstSolValue;
/* numbert of availabel solutions */
extern int gNumSolutions;
/* names and functions for heuristics used */
extern t_heuristic heuristic[];
/* diz se o A-Teams deve parar */
extern int gStopAteams;
/* number of iteractions */
extern int gNumIter; 
/* max number of iterations                     */
extern int g_max_num_iter;
/* say if there is a limit to number of iterations */
extern int gIterLimit;
/* pointers for list of results */
extern result_t *result, *last_res;
/* pointers to shortest paths */
extern int **A, **Pi;

/* defined in wmain.c */
#ifdef _WINDOWS_
/* faz a sincronizacao entre threads */
extern HANDLE gMutex, gMutexStop;
/* diz se o A-Teams ja comecou a rodar */
extern int gATRunning;
#endif /* _WINDOWS_ */


#endif /* _ATEAMS_H */

