/************************************************************************
 * grafo.c
 * Implements the graph abstract data type.
 *
 * (c) Carlos Oliveira, 1995-1999
 ***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <assert.h>

#include "graph.h"

/* Memory buffer to allocate graph_node */
typedef struct {
  graph_node *lista;
  struct blocks *block;	/* list of alocated blocks */
} memBuffer;			


/* a block of graph_nodes */
struct blocks {
  graph_node *start;
  struct blocks *prox;
};

/* Define a static memory buffer to graph_node */
static memBuffer buffer;

/* Size of the memory used (number of solutions in main memory */
static long int gn_buff_size;


/* Initialize the linked list of graph_nodes */
static void init_graph_node_buffer()
{
  int i;
  gn_buff_size = 512;
  if (!(buffer.lista = NewVec(graph_node, gn_buff_size)) ){
    printf("no available memory for graph_node!\n");
    exit(1);
  }

  /* links all of the graph_nodes together */
  for (i=0; i<gn_buff_size-1; i++) {
    buffer.lista[i].prox = &buffer.lista[i+1];
  }

  buffer.lista[gn_buff_size-1].prox = NULL;
  /* aloca memoria para a lista de blocos */
  buffer.block = New(struct blocks);
  if (!buffer.block) {
    printf("Falta memoria para lista de blocos");
    exit(1);
  }
  buffer.block->start = buffer.lista;
  buffer.block->prox = NULL;
}

/* Gets a new graph_node from the memory pool.
   The graph_nodes are allocated in a block, placed in buffer.lista.
   If the block is empty, a new one is alocated, and placed in a 
   linked list. 
   The linked list is used to delete all the graph_nodes allocated at
   the end of the program.
 */
graph_node *new_graph_node()
{
  graph_node *mem;
  static int flag=1;
  int i;
  struct blocks *b;

  if (flag) {
    /* Make the initialization of the blocks. It's called
       only at the first use of the function */
    init_graph_node_buffer(); 
    flag=0; 
  }
  /* the list of graph_nodes is exausted? */
  if (!buffer.lista) {
    /* alloc more memory */
    buffer.lista = NewVec(graph_node, gn_buff_size);
    if ( !buffer.lista ) {
      printf("no memory to new graph_node\n");
      exit(1);
    }
    /* link all the graph_nodes of the block */
    for (i=0; i<gn_buff_size-1; i++)
      buffer.lista[i].prox = &buffer.lista[i+1];
    buffer.lista[gn_buff_size - 1].prox = NULL;
    /* encadeia o novo bloco encontrado na lista de blocos alocados */
    b = New(struct blocks);
    if (!b) {
      printf("no memory to new buffer of graph_node\n");
      exit(1);
    }
    b->start = buffer.lista;
    b->prox = buffer.block;
    buffer.block = b;
  }
  mem = buffer.lista;
  buffer.lista = buffer.lista->prox;
  mem->prox = NULL;
  return mem;
}

/* Returns the graph_node to the memory pool */
void delete_graph_node(graph_node *l)
{
  l->prox = buffer.lista;
  buffer.lista = l;
}

/*
 * graph implementation using foward-star.
 */
#ifdef FORWARDSTAR
 
Graph *grfs_create (int nVert)
{
  Graph *G;
  NewS(G, Graph, e0);
  NewVec(G->v, int, nVert+2, e1);
  NewVec(G->a, int, 1000, e2);
  NewVec(G->w, int, 1000, e3);

  G->nVerts  = nVert;
  G->nArSize = 1000;
  G->nArests = 0;
  memset(G->v, 0, sizeof(int)*(nVert+2));
  return G;

 e3: Delete(G->a);
 e2: Delete(G->v);
 e1: Delete(G);
 e0: return 0;
}

void grfs_delete (Graph * G)
{
  Delete(G->v);
  Delete(G->a);
  Delete(G->w);
  Delete(G);
}

Graph *grfs_copy (Graph * G)
{
  int i;
  Graph *Gn;
  NewS(Gn, Graph, e0);
  NewVecS(Gn->v, int, G->nVerts+2, e1);
  NewVecS(Gn->a, int, G->nArSize, e2);
  NewVecS(Gn->w, int, G->nArSize, e3);

  Gn->nVerts  = G->nVerts;
  Gn->nArSize = G->nArSize;
  Gn->nArests = G->nArests;
  for (i = 1; i <= G->nVerts+1; i++)
    Gn->v[i] = G->v[i];
  for (i = 1; i <= G->nArests; i++){
    Gn->a[i] = G->a[i];
    Gn->w[i] = G->w[i];
  }
  return Gn;
  
 e3: Delete(Gn->a);
 e2: Delete(Gn->v);
 e1: Delete(Gn);
 e0: return 0;
}

/* Set arests in a directed graph.
   Input: A pointer to a Graph structure, the two linked vertices,
   the weight between them.
   
   Insert the edge in a linked list of vertices adjacent to source,
   ordered by destination vertex number */
int grfs_set_arest_dir(Graph * G, vert a, vert b, int cost)
{
  int t, p;

  /* find the correct place to insert */
  for (p = G->v[a]; G->a[p] <= b && p < G->v[a+1];)
    p++;

  for (t= 2*(G->nArests+1); t > p; t--){
    G->a[t] = G->a[t-1];
    G->w[t] = G->w[t-1];
  }
  G->a[t] = b;
  G->w[t] = cost;
  for (t= a+1; t <= G->nVerts+1; t++)
    G->v[t]++;
  return cost;
}

/* set an arest between a and b with weight `cost' */
int grfs_set_arest(Graph * G, vert a, vert b, int cost)
{
  if (a==b)
    return 0;

  if ((2*G->nArests) >= (G->nArSize-1)){
    /* aloca mais memoria para arestas */
    if (!(G->a = realloc(G->a, 2*G->nArests + 128)))
      return 0;
    if (!(G->w = realloc(G->w, 2*G->nArests + 128)))
      return 0;
  }

  if (grfs_set_arest1(G, a, b, cost)
      && grfs_set_arest1(G, b, a, cost))
    G->nArests++;
  else 
    return 0;
  return cost;
}


/* returns the weight of an arest */
int grfs_get_cost(Graph * G, vert a, vert b)
{
  int i, j, *x;

  j = G->v[a+1] - G->v[a];
  x = &G->a[G->v[a]];
  for (i=0; i<j && *x<b; i++, x++)
    ;
  if (i<j && *x == b)
    return G->w[G->v[a]+i];
  return 0;

}

int grfs_get_total_cost(Graph * G)
{
  int i, n = gr_nverts(G), *x, *y, *start, sum = 0;
  start = G->a;
  for (i=1; i<=n; i++) {
    x = &G->a[G->v[i]];
    y = &G->a[G->v[i+1]];
    for (; x < y && *x < i; x++) {
      sum += G->w[start - x];
    }
  }
  return sum;
}

int grfs_vdegree(Graph * G, vert v)
{
  if (v < 1 || v > gr_nverts(G))
    return 0;
  return (G->v[v+1] - G->v[v]);
}

graph_node *grfs_get_adj_list(Graph *G, vert v)
{
  graph_node *h, *l;
  int d = gr_vdegree(G, v);
  int i;
  h = l = new_graph_node();
  if (!h)
    return 0;
  h->v = d;
  for (i=0; i<d; i++){
    l->prox = new_graph_node();
    if (!l->prox)
      goto error;
    l = l->prox;
    l->v     = G->a[G->v[v]+i];
    l->custo = G->w[G->v[v]+i];
    l->prox  = 0;
  }
  return h;

  error:
    while (h){
      l = h->prox;
      delete_graph_node(h);
      h = l;
    }
    return 0;
}

void grfs_remove_all_edges(Graph *G)
{
  gr_nverts(G) = 0;
  gr_narests(G) = 0;
}

void grfs_print(Graph *G)
{
  int i, j, n;
  n = gr_nverts(G);
  for (i=1; i<=n; i++){
    if (G->v[i] != G->v[i+1]) {
      printf("%d => ", i);
    }
    for (j=G->v[i]; j<G->v[i+1]; j++){
      printf("%d ", G->a[j]);
    }
    if (G->v[i] != G->v[i+1]) {
      printf("\n");
    }
  }
}


#endif /* FORWARDSTAR */



/*
 * graph implementation using adjacency list
 */
#ifdef	ADJLIST

Graph *gral_create(int n)
{
  int i;
  Graph *G;

  NewS(G, Graph, e1);
  NewVecS(G->node, graph_node, n+1, e2);
  for (i = 0; i <= n; i++){
    G->node[i].custo = 0;   /* store the number of arests        */
    G->node[i].prox  = 0;
    G->node[i].v     = 0;
  }
  G->node[0].v = n;         /* number of vertices */
  G->A = 0;
  gr_set_graph_matrix(G);
  return G;

 e2: Delete(G->node);
 e1: return 0;
}

void gral_delete (Graph * G)
{
  int i, n = gr_nverts (G); 
  for (i = 0; i <= n; i++) {
    graph_node *no = G->node[i].prox;
    while (no) {
      /* deleta a lista de adjacencias */
      graph_node *aux = no;
      no = no->prox;
      delete_graph_node(aux);
    }
  }
  if (G->A) {
    DelVec2(G->A, n+1);
  }
  Delete(G->node);
  Delete(G);
}

Graph *gral_copy (Graph * G)
{
  int i, n = gr_nverts(G);
  Graph *Copy = gr_create(n);
  graph_node *no, *no2;

  Copy->node[0].v = G->node[0].v;
  for (i = 0; i<=n; i++) {
    Copy->node[i].custo = G->node[i].custo;
    Copy->node[i].prox = 0;
    for (no = G->node[i].prox, no2 = &Copy->node[i]; no; no=no->prox) {
      no2->prox  = new_graph_node();
      no2        = no2->prox;
      no2->v     = no->v;
      no2->custo = no->custo;
      no2->prox  = 0;
    }
  }
  if (G->A) {
    gr_set_graph_matrix(Copy);
  }
  return Copy;
}

int gral_set_graph_matrix(Graph *G)
{
  int i, n = gr_nverts(G);
  graph_node *node;

  if (!G->A) {
    NewVec2Szero(G->A, int, n+1, n+1, error);
    for (i=1; i<=n; i++) {
      for (node=G->node[i].prox; node; node=node->prox) {
	G->A[i][node->v] = node->custo;
      }
    }
  }
  return 1;

 error: return 0;
}

/* Set arests in a directed graph. Adjacencies are stored in nondecreasing
   order.
   Input:
   ------
   G: pointer to the Graph structure;
   a, b: the two linked vertices;
   cost:the weight between them.
   Return:
   -------
   1 when successful
   2 when arest already exists
   0 on failure. */
int gral_set_arest_dir(Graph *G, vert a, vert b, int cost)
{
  graph_node *no = &G->node[a], *aux;
  while (no->prox && no->prox->v <= b)
    no = no->prox;
  
  if (no->v == b && no != &G->node[a]) {
    /* we only change the cost */
    no->custo = cost; 
    return 2; 
  }
  
  /* insert in order */
  aux = no->prox;
  no->prox = new_graph_node();
  if (!no->prox)
    return 0;
    
  no->prox->v = b;
  no->prox->custo = cost;
  no->prox->prox = aux;
  G->node[a].custo++;
  if (G->A) {
    G->A[a][b] = cost;
  }
  return 1;
}

int gral_set_arest(Graph *G, vert a, vert b, int cost)
{
  int res1, res2;
  res1 = gr_set_arest_dir(G, a, b, cost);
  res2 = gr_set_arest_dir(G, b, a, cost);
  /* if no error was returned, increase the arest number */
  if ( res1 != 2 && res2 != 2 && res1 != 0 && res2 != 0 ) {
    G->node[0].custo++;
    return a + b;
  } else {
    return 0;
  }
}

int gral_remove_edge_dir(Graph *G, vert a, vert b)
{
  int nv, changed = 0;
  graph_node *n, *t;
  nv = gr_nverts(G);
  assert (a > 0 && a <= nv);
  for (n = &G->node[a]; n->prox;) {
    if (n->prox->v == b) {
      t = n->prox;
      n->prox = t->prox;
      delete_graph_node(t);
      changed = 1;
    } else {
      n = n->prox;
    }
  }
  if (changed) {
    G->node[a].custo--;
    if (G->A) {
      G->A[a][b] = 0;
    }
  }
  return changed;
}

int gral_remove_edge(Graph *G, vert a, vert b)
{
  int changed = 0;
  changed = gral_remove_edge_dir(G, a, b);
  if (changed) {
    gral_remove_edge_dir(G, b, a);
    G->node[0].custo--;
  }
  return changed;
}

/* Funcao que retorna o custo de uma aresta */
int gral_get_cost (Graph *G, vert a, vert b)
{
  graph_node *no;
  int n;
  n = gr_nverts(G);
  assert(a > 0 && a <= n && b > 0 && b <= n);
  if (G->A) {
    return G->A[a][b];
  }
  if ( a < b ) {
    swap(vert, a, b);
  }
  no = G->node[a].prox;
  while (no && no->v < b)
    no = no->prox;
  if (no && no->v == b) {
    return no->custo;
  } else {
    return 0;
  }
}

int gral_get_total_cost(Graph *G)
{
  int i, sum = 0, n = gr_nverts(G);
  graph_node *node;
  for (i=1; i<=n; i++) {
    for (node = G->node[i].prox; node; node=node->prox) {
      sum += node->custo;
    }
  }
  return sum;
}

/* create a list of adjacencies to the vertex v. The header of the
   list stores the # of elements */
graph_node *gral_get_adj_list(Graph *G, vert v)
{
  graph_node *h, *l, *vl;
  h = l = new_graph_node();
  if (!h) return 0;
  l->prox = 0;
  l->v = gr_vdegree(G, v);
  for (vl = G->node[v].prox; vl; vl=vl->prox){
    l->prox = new_graph_node();
    l = l->prox;
    if (!l)
      goto error;
    l->v     = vl->v;
    l->custo = vl->custo;
    l->prox  = 0;
  }
  return h;

  error:
    while (h){
      l = h->prox;
      delete_graph_node(h);
      h = l;
    }
    return 0;  
}

void delete_list_of_nodes(graph_node *h)
{
  graph_node *no;
  while (h){
    no = h->prox;
    delete_graph_node(h);
    h = no;
  }
}

void gral_print(Graph *G)
{
  int i, n;
  graph_node *no;
  n = gr_nverts(G);
  for (i=1; i<=n; i++){
    no = G->node[i].prox;
    while (no){
      printf("%d ", no->v);
      no = no->prox;
    }
    printf("\n");
  }
}

#endif /* ADJLIST */

/* Funcao que le um grafo a partir de um arquivo externo */
Graph *read_graph_from_file(FILE * f)
{
  int n, na, i;
  Graph *G;
  char error[] = "Error in graph description file.\n";
  char c1[10], c2[10], c3[10], c4[10];
  /* read the number of vertex */
  fscanf (f, "%s %s", c1, c2);
  if (!strcmp(c1, "n") && !strcmp(c1, "N")) {
    puts(error);
    return 0;
  }
  n = atoi(c2);
  if (!(G = gr_create(n)))
      return 0;
  fscanf (f, "%s %s", c1, c2);
  if (strcmp (c1, "m") && strcmp (c1, "M")) {
    puts (error);
    return 0;
  }
  na = atoi(c2);
  for (i = 0; i < na; i++) {
      /* le cada aresta */
      int a, b, c;
      fscanf (f, "%s %s %s %s", c1, c2, c3, c4);
      if (strcmp (c1, "a")) {
	puts(error);
	return 0;
      }
      a = atoi(c2);
      b = atoi(c3);
      c = atoi(c4);
      if (!gr_set_arest (G, a, b, c))
	return 0;
    }
  return G;
}

int gr_forest_cost(Graph *G)
{
  int sum = 0, n = gr_nverts(G), i;
  graph_node *no;
  
  for (i = 1; i <= n; i++) {
    for (no=G->node[i].prox; no; no=no->prox) {
      if (i < no->v) {
	sum += no->custo;
      }
    }
  }
  return sum;
}

void gr_return_adj_list(graph_node *n)
{
  delete_list_of_nodes(n);
}

void gral_remove_all_edges(Graph *G) 
{
  int n = gr_nverts(G);
  int i;
  graph_node *node, *t;
  for (i=1; i<=n; i++){
    for (node = G->node[i].prox; node;){
      t = node->prox;
      delete_graph_node(node);
      node = t;
    }
    G->node[i].prox = 0;
    G->node[i].custo = 0;
    G->node[i].v = 0;
  }
  gr_narests(G) = 0;
  if (G->A) {
    for (i=0; i<=n; i++) {
      memset(G->A[i], 0, sizeof(int)*(n+1));
    }
  }
}
