/*----------------------------------------------------------------
 *  ATeams
 *  Program to solve the Point-to-point connection problem
 *
 *  (C)Carlos A.S. Oliveira 1996-2001
 *----------------------------------------------------------------*/

/* heurist.c: implementation of heuristics */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include "common.h"
#include "ateams.h"

/* variables used in this module only */
static int *s, *cs, *cd;
static int *used, *sa, *sb;
static arest *ar1, *ar2, *intersec, *dif;

/* inicial memory alocation */
void InitHeurist()
{
  int p = gInst->p;
  int nArest = gr_narests(gInst->G);
  int nVert = gr_nverts(gInst->G);

  NewVecS(cs, int, gInst->p, erro1);
  NewVecS(cd, int, gInst->p, erro2);
  NewVecS(s,  int, p, erro3);
  NewVecS(ar1, arest, nArest, erro4);
  NewVecS(ar2, arest, nArest, erro5);
  NewVecS(intersec, arest, nArest, erro6);
  NewVecS(used, int, nArest, erro7);
  NewVecS(dif, arest, nArest, erro8);
  NewVecS(sa, int, nVert, erro9);
  NewVecS(sb, int, nVert, erro10);
  return;

 erro10: Delete(sa);
 erro9:  Delete(dif);
 erro8:  Delete(used);
 erro7:  Delete(intersec);
 erro6:  Delete(ar2);
 erro5:  Delete(ar1);
 erro4:  Delete(s);
 erro3:  Delete(cd);
 erro2:  Delete(cs);
 erro1:  fatal_error("Error: Could not alloc memory in Init Heurist!\n");
}

void DeleteHeurist()
{
  Delete(sb);
  Delete(sa);
  Delete(dif);
  Delete(used);
  Delete(intersec);
  Delete(ar2);
  Delete(ar1);
  Delete(s);
  Delete(cd);
  Delete(cs);
}

/* creates a completely random solution */
Solution *Heuristic1 (Graph * G, vert s[], vert d[], int p)
{
  int i, dpos;
  Solution *sol;

  NewVecS(sol, Solution, p, e0);
  memset(cd, 0, sizeof(int)*p);
  for (i = 0; i < p; i++){
    /* find a random value for destination */
    do {
      dpos = random_int(p);
    } while (cd[dpos] != 0);
    cd[dpos] = 1;

    /* create a random path */
    sol[i].dest = d[dpos];
    sol[i].path = random_path(G, s[i], d[dpos]);
    if (!sol[i].path) goto e1;
  }
  return sol;

 e1:
  for (i--; i>=0; i--) {
    delete_path(sol[i].path);
  }
 e0: return 0;
}

/* improvement heuristic I1: uses the edge with lowest weight to
    concatenate shortest paths in a solution. */
Solution *Heuristic5(Graph *G, vert source[], vert dest[], int p)
{
  int npath, min, custo;
  vert a, b, s, d;
  listvert *l;
  Solution *sol, *s_orig;

  /* select a solution */
  s_orig = solucoes[random_int(gNumSolutions)].sol;
  sol = copy_solution(s_orig, p);
  /* select a path in the solution */
  npath = random_int(p);
  l = sol[npath].path;
  a = s = source[npath];
  d = sol[npath].dest;
  min = gr_get_cost(G, l->v, l->prox->v);
  l = l->prox;
  b = l->v;
  /* find the edge with minimum cost in this path */
  if (l)
    while (l->prox) {
      if ((custo = gr_get_cost (G, l->v, l->prox->v)) < min) {
	min = custo;
	a = l->v;
	b = l->prox->v;
      }
      l = l->prox;
    }
  l = sol[npath].path;
  /* delete the old path */
  while (l) {
    listvert *aux = l;
    l = l->prox;
    delete_list_vert(aux);
  }
  /* construct two shortest paths, between a-u and v-b */
  if (s != a) {
    l = sol[npath].path = shortest_path (G, s, a);
    while (l->prox)
      l = l->prox;
  } else {
    l = sol[npath].path = new_list_vert();
    l->v = a;
  }
  if (b != d) {
    l->prox = shortest_path (G, b, d);
  } else {
    l->prox = new_list_vert();
    l->prox->v = b;
    l->prox->prox = NULL;
  }
  /* remove cycles in this path */
  delete_loop(sol[npath].path, G);
  return sol;
}

#if 0
Solution *permute_paths (Solution *sol, Graph * G, int *s, int p)
{
  int a, b, l;
  listvert *aux1, *aux2;

  a = random_int(p);
  do {
    b = random_int(p);
  } while (b == a);

  l = sol[a].dest;
  /* Deleta a lista de vertices pertencentes a sol[a] */
  aux1 = sol[a].path;
  while (aux1) {
    aux2 = aux1;
    aux1 = aux1->prox;
    delete_list_vert(aux2);
  }
  /* Gera um novo caminho para sol[a] */
  sol[a].path = random_path (G, s[a], sol[b].dest);
  sol[a].dest = sol[b].dest;
  /* Deleta a lista de vertices pertencentes a sol[a] */
  aux1 = sol[b].path;
  while (aux1) {
    aux2 = aux1;
    aux1 = aux1->prox;
    delete_list_vert(aux2);
  }
  /* Gera um novo caminho para sol[b] */
  sol[b].path = random_path (G, s[b], l);
  sol[b].dest = l;
  return sol;
}
#endif


/* improvement heuristic I2: find the weightest arest in one of the
   paths of the solution and try to find a better path between its
   vertices */
Solution *Heuristic6(Graph *G, vert source[], vert dest[], int p)
{
  int i, npath, max, custo, randSol = random_int(gNumSolutions);
  vert a, b;
  listvert *l1, *l2, *sp;
  Solution *sol = copy_solution(solucoes[randSol].sol, p);

  memset(s, 0, sizeof(int)*p);
  for (i = 0; i < p; i++) {
    /* find a random path in the solution */
    do {
      npath = random_int(p);
    } while (s[npath] == 1);
    s[npath] = 1;
    max = 0;
    l1 = sol[npath].path;
    a = l1->v;
    b = l1->prox->v;
    /* Find the arest with the maximum weight in this
       path to improve it */
    while (l1->prox) {
      if ((custo = gr_get_cost (G, l1->v, l1->prox->v)) > max) {
	max = custo;
	a = l1->v;
	b = l1->prox->v;
	l2 = l1;		/* keep pointer to the 1st node */
      }
      l1 = l1->prox;
    }
    l1 = sp = shortest_path (G, a, b);
    custo = 0;
    /* now we calculate the cost of this paths */
    while (l1->prox) {
      custo += gr_get_cost (G, l1->v, l1->prox->v);
      l1 = l1->prox;
    }
    if (custo >= max) {
      /* don't change sol */
      while (sp) {
	/* delete sp */
	listvert *aux = sp;
	sp = sp->prox;
	delete_list_vert(aux);
      }
      continue; /* try another path */
    }
    /* concat the paths */
    l1->prox = l2->prox->prox;
    delete_list_vert(l2->prox);	/* delete the extra vertex ( b ) */
    l2->prox = sp->prox;
    delete_list_vert(sp);	/* delete the extra vertex ( a ) */
    

    /* remove cycles in the path */
    delete_loop (sol[i].path, G);
    return sol;
  }
  /* if no path could be improved return the same solution */
  return sol;
}


/* This heuristic works with heuristic #9 and fills the
   PartialMem memory with some of the common arests of two 
   solutions from the shared memory */
static void Heuristic2_partial(Graph * G, int p)
{
  int nsol1, nsol2, i, j, top1, top2, top3 = -1;
  int nArest = gr_narests(G), cont = -1;
  Solution *sol1, *sol2;

  for (i = 0; i < nArest; i++)
    used[i] = 0;
  /* Generate 2 random solution numbers in M1 */
  nsol1 = gNumSolutions - linear_rand(gNumSolutions - 1);
  do {
    nsol2 = gNumSolutions - linear_rand(gNumSolutions - 1);
  } while (nsol1 == nsol2);
  sol1 = solucoes[nsol1].sol;
  sol2 = solucoes[nsol2].sol;

  /* Find the arests that exists in each solution */
  top1 = arests_in_solution(sol1, ar1, p);
  top2 = arests_in_solution(sol2, ar2, p);

  /* set number of elements in PartialMem to 0 */
  PartialMem.n = 0;

  /* Iterate to each arest in ar1 and ar2 */
  for (i = 0; i <= top1; i++)
    for (j = 0; j <= top2; j++)
      /* Copy only  common arests in ar1 and ar2 */
      if ((ar1[i].a == ar2[j].a && ar1[i].b == ar2[j].b) ||
	  (ar1[i].b == ar2[j].a && ar1[i].a == ar2[j].b))
	{
	  intersec[++top3].a = ar1[i].a;
	  intersec[top3].b = ar1[i].b;
	}

  for (i = 0, cont = top3; i < p; i++) {
    int n;
    /* if insersec is empty , then break */
    if (cont-- == -1)
      break;
    do {
      n = random_int(top3 + 1);
    } while (used[n] == 1);
    used[n] = 1;
    PartialMem.ar[i].a = intersec[n].a;
    PartialMem.ar[i].b = intersec[n].b;
  }
  PartialMem.n = i;
}


Solution *Heuristic2(Graph * G, vert s[], vert d[], int p)
{
  Heuristic2_partial(G, p);
  return constructor2(G, s, d, p);
}


/* This heuristics works with heuristic #9 and it fills the
   PartialMem memory with some arests that exists only in one 
   solution better then other choosed randomically in 
   the shared memory */
static void Heuristic3_partial(Graph * G, int p)
{
  int randSol1 = random_int(gNumSolutions);
  int randSol2, top1 = -1, top2 = -1, top3 = -1;
  int nArest = gr_narests(G), custo1, custo2, i, j, cont;
  Solution *sol1, *sol2;

  /* Fill used with zeros (none arest choosed) */
  for (i = 0; i < nArest; i++)
    used[i] = 0;
  /*  Choose two randomic solutions            */
  do {
    randSol2 = random_int(gNumSolutions);
  }
  while (randSol1 == randSol2);
  sol1 = solucoes[randSol1].sol;
  sol2 = solucoes[randSol2].sol;
  top1 = arests_in_solution(sol1, ar1, p);
  top2 = arests_in_solution(sol2, ar2, p);
  custo1 = eval_solution(sol1, G, p);
  custo2 = eval_solution(sol2, G, p);
  if (custo1 < custo2)
    for (i = 0; i <= top1; i++) {
      int flag = 0;
      for (j = 0; j <= top2; j++)
	if ((ar1[i].a == ar2[j].a && ar1[i].b == ar2[j].b) ||
	    (ar1[i].a == ar2[j].b && ar1[i].b == ar2[j].a))
	  {
	    flag++;
	    break;
	  }
      /* Copy ar1[i] only if it's not equal to a arest in ar2 */
      if (!flag) {
	dif[++top3].a = ar1[i].a;
	dif[top3].b = ar1[i].b;
      }
    } else
      for (i = 0; i <= top2; i++) {
	int flag = 0;
	for (j = 0; j <= top1; j++)
	  if ((ar2[i].a == ar1[j].a && ar2[i].b == ar1[j].b) ||
	      (ar2[i].a == ar1[j].b && ar2[i].b == ar1[j].a))
	    {
	      flag++;
	      break;
	    }
	if (!flag)
	  dif[++top3] = ar2[i];
      }
  /* Put the arests in PartialMem */
  for (i = 0, cont = top3; i < p; i++) {
    int randomPos;
    /* if insersec is empty , then break */
    if (cont-- == -1)
      break;
    do {
      randomPos = random_int(top3 + 1);
    } while (used[randomPos] == 1);
    used[randomPos] = 1;
    PartialMem.ar[i].a = dif[randomPos].a;
    PartialMem.ar[i].b = dif[randomPos].b;
  }
  PartialMem.n = i;
}

Solution *Heuristic3(Graph * G, vert s[], vert d[], int p)
{
  Heuristic3_partial(G,p);
  return constructor2(G, s, d, p);
}


/* This heuristic takes some edges in the Partial memory
   and create a new solution.
   Algorithm:
   for (each edge (i,j) in partial mem)
     choose a source node a (can be one of (i,j))
     choose a destination node b (can also be one of (i,j))
     find the paths (shortest-path(a-i) and shortest-path(j-b))
   if there are still unmatched pairs
     insert random paths   
*/
Solution *constructor2(Graph * G, vert source[], vert dest[], int p)
{
  int i, a, b, nVert = gr_nverts(G);
  listvert *l;
  Solution *sol;

  sol = NewVec (Solution, p);
  if (!sol) {
    fatal_error("not enough memory on constructor 2\n");
  }
  memset(sa, 0, sizeof(sa[0])*nVert);
  memset(sb, 0, sizeof(sa[0])*nVert);
  for (i = 0; i < PartialMem.n; i++) {
    int j, flaga = 0, flagb = 0;
    /* choose a source node */
    for (j = 0; j < p; j++) {
      if (source[j] == PartialMem.ar[i].a && sa[j] == 0) {
	flaga++;
	break;
      }
    }
    if (flaga) {
      a = j;
    } else {
      do {
	a = random_int(p);
      } while (sa[a] == 1);
    }
    /* choose a destination node */
    for (j = 0; j < p; j++) {
      if (dest[j] == PartialMem.ar[i].b && sb[j] == 0) {
	flagb++;
	break;
      }
    }
    if (flagb) {
      b = j;
    } else {
      do {
	b = random_int(p);
      } while (sb[b] == 1);
    }
    sa[a] = sb[b] = 1; /* mark vertices a and b */
    
    /* Generate a path from a to first vertex in PatialMem.ar[i]  */
    if (!flaga) {
      l = sol[a].path = shortest_path (G, source[a], PartialMem.ar[i].a);
    } else {
      l = sol[a].path = new_list_vert();
      l->v = PartialMem.ar[i].a;
      l->prox = NULL;
    }
    while (l->prox) {
      l = l->prox;
    }
    /* Generate a path from the second vertex of the arest we're using
       to the destination b */
    if (!flagb) {
      l->prox = shortest_path (G, PartialMem.ar[i].b, dest[b]);
      sol[a].dest = dest[b];
    } else {
      l->prox = new_list_vert();
      l->prox->v = PartialMem.ar[i].b;
      l->prox->prox = NULL;
      sol[a].dest = PartialMem.ar[i].b;
    }
    }
  /* If the solution sol isn't complete yet, use ramdom paths
     between the vertex source and destination  */
  for (; i < p; i++) {
    do {
      a = random_int(p);
    } while (sa[a] == 1);
    do {
      b = random_int (p);
    }
    while (sb[b] == 1);
    sa[a] = sb[b] = 1;
    sol[a].path = random_path (G, source[a], dest[b]);
    sol[a].dest = dest[b];
  }
  for (i = 0; i < p; i++)
    delete_loop(sol[i].path, G);
  return sol;
}

/* Creates a new solution based in shortest paths */
Solution *Heuristic8(Graph * G, vert source[], vert dest[], int p)
{
  int randSol = random_int (gNumSolutions), i;
  Solution *sol, *aux = solucoes[randSol].sol;

  if (!(sol = NewVec (Solution, p)))
    fatal_error ("Coudn't alloc memory in Heurist10!\n");
  for (i = 0; i < p; i++) {
    sol[i].dest = aux[i].dest;
    sol[i].path = shortest_path(G, source[i], aux[i].dest);
  }
  return sol;
}

/* Heuristica da Desigualdade triangular */
Solution *Heuristic7(Graph * G, vert source[], vert dest[], int p)
{
  int randSol = random_int(gNumSolutions), i;
  int a, b, c;
  int custo, soma;
  Solution *sol = copy_solution(solucoes[randSol].sol, p);

  /* Search in each path of the solution */
  for (i = 0; i < p; i++)  {
    listvert *l = sol[i].path;
    if (l && l->prox)
      for (; l->prox->prox; l = l->prox) {
	a = l->v;
	b = l->prox->v;
	c = l->prox->prox->v;
	custo = gr_get_cost (G, a, c);
	if (custo == 0)
	  continue;
	soma = gr_get_cost (G, a, b) + gr_get_cost (G, b, c);
	/* Delete the central vertex */
	if (custo < soma) {
	  listvert *temp = l->prox;
	  l->prox = l->prox->prox;
	  delete_list_vert(temp);
	  if (!(l->prox->prox))
	    break;
	}
      }
  }
  return sol;
}

Solution *Heuristic4(Graph *G, vert s[], vert d[], int p)
{
  int i, j, spos;
  Solution *sol;
  listvert *temp;

  if (!(sol = NewVec (Solution, p))) {
    fatal_error("Coudn't alloc memory in C3!\n");
  }
  for (i = 0; i < p; i++){
    cs[i] = s[i];
    cd[i] = d[i];
  }
  for (i = 0; i < p; i++){
    vert v, dest;
    do	{
      spos = random_int (p);
    } while (cs[spos] == -1);
    v = s[spos];
    temp = random_path_sdfree(G, &v, &dest, cs, cd, p);

    for (j = 0; j < p; j++){
      if (cd[j] == dest){
	cd[j] = -1;
	break;
      }
    }
    for (j = 0; j < p; j++){
      if (cs[j] == v){
	cs[j] = -1;
	spos = j;
	break;
      }
    }
    sol[spos].path = temp;
    sol[spos].dest = dest;
  }
  return sol;
}

/* given the pairs (a,c) and (b,d) we find a way of merge the
   paths from a to c into that of b to d. */
Solution *join_paths(Solution *sol, Graph * G, int *s, int p, 
		     int a, int b)
{
  int d, db=9999, dd=9999, vb, vd, k, last;
  listvert *l1, *l2, *l3, *lb, *ld;

  d = sol[b].dest;
  /* find the vertex with the least distance to b and d */
  for (l1=sol[a].path; l1; l1=l1->prox) {
    k = shortest_path_dist(G, l1->v, s[b]);
    if (k < db) {
      db = k;
      vb = l1->v;
      lb = l1;
      last = 1;
    }
    k = shortest_path_dist(G, l1->v, d);
    if (k < dd) {
      dd = k;
      vd = l1->v;
      ld = l1;
      last = 2;
    }
  }
  /* Delete the list of vertices from s[b] to d */
  l1 = sol[b].path;
  while (l1) {
    l2 = l1;
    l1 = l1->prox;
    delete_list_vert(l2);
  }
  sol[b].path = l1 = sol[b].path = shortest_path(G, s[b], vb);
  while (l1->prox)
    l1=l1->prox;
  if (last == 1) {
    /* we must revert the path */
    l3 = l2 = shortest_path(G, vd, d);
    l2 = l2->prox;
    delete_list_vert(l3);
    while (ld->v != vb){
      l3 = new_list_vert();
      l3->v = ld->v;
      l3->prox = l2;
      l2 = l3;
      ld = ld->prox;
    }
    l1->prox = l2;
  } else {
    if (lb->v != vd) {
      for (lb=lb->prox; lb->v!=vd; lb=lb->prox) {
	l1->prox = new_list_vert();
	l1 = l1->prox;
	l1->v = lb->v;
      }
      l1->prox = new_list_vert();
      l1 = l1->prox;
      l1->v = vd;
    }
    l1->prox = shortest_path(G, vd, d);
    l2 = l1->prox;
    l1->prox = l1->prox->prox;
    delete_list_vert(l2);
  }
  return sol;
}

Solution *Heuristic9(Graph *G, vert s[], vert d[], int p)
{
  int a, b, pos = random_int(gNumSolutions);
  Solution *sol = copy_solution(solucoes[pos].sol, p);
  a = random_int(p);
  do {
    b = random_int(p);
  } while (a == b);
  sol = join_paths(sol, G, s, p, a, b);
  delete_loop(sol[b].path, G);
  return sol;
}

/* Select two pais (a,c), (b,d) and try to merge the path with 
   best "arest usage" value with the other */
Solution *Heuristic10(Graph *G, vert s[], vert d[], int p)
{
  int a, b, u, v, sum_a, sum_b, pos = random_int(gNumSolutions);
  listvert *l;
  Graph *used_edges;
  Solution *sol = copy_solution(solucoes[pos].sol, p);
  used_edges = gr_create(gr_nverts(G));
  calc_arest_usage(G, used_edges, sol, p);
  a = random_int (p);
  do {
    b = random_int (p);
  } while (b == a);

  /* calculate the sum of arest usage for first path */
  for (l=sol[a].path; l->prox; l=l->prox) {
    u = l->v;
    v = l->prox->v;
    sum_a += gr_get_cost(used_edges, u, v);
  }
  /* calculate the sum of arest usage for second path */
  for (l=sol[b].path; l->prox; l=l->prox) {
    u = l->v;
    v = l->prox->v;
    sum_b += gr_get_cost(used_edges, u, v);
  }

  if (sum_a > sum_b) {
    sol = join_paths(sol, G, s, p, a, b);
    delete_loop(sol[b].path, G);
  } else {
    sol = join_paths(sol, G, s, p, b, a);
    delete_loop(sol[a].path, G);
  }
  gr_delete(used_edges);
  return sol;
}


Solution *Heuristic11(Graph *G, vert s[], vert d[], int p)
{
  float lb;
  static Graph *H = 0;
  static restricted_forest_tad *rf;

  if (!H) {
    rf = rf_create(gInst, rf_f);
    H = restricted_forest(rf, &lb);
  }
  return rf_create_solution_from_graph(rf, H);
}


Solution *Heuristic12(Graph *G, vert s[], vert d[], int p)
{
  int n;
  Graph *H;
  Solution *sol;
  restricted_forest_tad *rf;

  n = gr_nverts(G);
  Heuristic2_partial(G, p);
  rf = rf_create(gInst, rf_f);
  get_verts_from_arests(PartialMem.ar, rf->part_verts, n, PartialMem.n);
  H = restricted_forest_withpt(rf, rf->part_verts, PartialMem.n);
  sol = rf_create_solution_from_graph(rf, H);
  gr_delete(H);  
  rf_delete(rf);
  return sol;
}
