/*****************************************************************
 * mst.c
 * programa que gera a minimum spanning tree para um grafo dado,
 * retira as folhas nao pertencentes a S uniao D para gerar um
 * upper bound para o PPC A-Teams.
 *
 * (C) 1996-2000, Carlos Alberto de S. Oliveira
 *****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include "ateams.h"

static arest *arestas;
static int *pq, *dad=0;


/*******
 * minimum spanning tree algorithm implementation
 */
int mst_init(ppc_instance *i)
{
  int n  = gr_nverts(i->G);
  int m = gr_narests(i->G);

  /* allocate memory to the vectors used */
  NewVecS(dad, int, n+1, e0);
  NewVecS(pq, int, m+1, e1);
  NewVecS(arestas, arest, m+1, e2);
  return 1;

 e2:  Delete(pq);
 e1:  Delete(dad);
 e0:  return 0;
}

void mst_delete()
{
  Delete(arestas);
  Delete(pq);
  Delete(dad);
}

/*********
 * Union-Find implementation
 *   dad is the set of integers stored
 *   returns true if a and b are not in the same set
 *   if doit is true, b is put in the same set as a.
 */
static int find (int dad[], int a, int b, int doit)
{
  int i = a, j = b;
  while (dad[i] > 0)
    i = dad[i];
  while (dad[j] > 0)
    j = dad[j];
  if (doit && (i != j))
    dad[j] = i;
  return (i != j);
}

/********
 * binary heap for minimum spanning tree implementation
 */
typedef struct {
  Graph *G;
  arest *ar;
} mst_heap_elem;

#define mstindex int
#define S ((mst_heap_elem *)s)
#define heap_compare(j,k) \
  (gr_get_cost(S->G, S->ar[a[(j)]].a, S->ar[a[(j)]].b) > \
   gr_get_cost(S->G, S->ar[a[(k)]].a, S->ar[a[(k)]].b) )
#define heap_swap(p,q)    { int  __x = a[p]; a[p]=a[q]; a[q]=__x; }
define_heap_operation(mstindex)
#undef heap_compare
#undef heap_swap
#undef S


/* find different arests in Graph G.
   returns the number of different arests found */
int get_arests_in_graph(Graph *G, arest *arestas)
{
  graph_node *h, *no;
  int i, j, n;

  assert(G);
  n = gr_nverts(G);
  
  for (i = 1, j = 1; i <= n; i++) {
    h = no = gr_get_adj_list(G, i);
    if (!no) continue;

    for ( no = no->prox; no && no->v < i; no=no->prox) {
      arestas[j].a = i;
      arestas[j++].b = no->v;
    }
    gr_return_adj_list(h);
  }
  return j-1;
}


/* Returns a graph representing a minimum spanning tree
   found by the kruskal method */
Graph *kruskal (Graph * G)
{
  Graph *Mst;
  int i, j, n = gr_nverts(G), m = gr_narests(G);
  int soma = 0, nElem;
  mst_heap_elem me;

  /* init the dad vector to zero */
  memset (dad, 0, sizeof (int) * (n+1));
  /* init the heap structure */
  for (i = 0; i<=m; i++) {
    pq[i] = i;
  }

  j = get_arests_in_graph(G, arestas);
  assert (j == m);
  /* Constructs the priority queue */
  me.G = G;
  me.ar = arestas;
  for (i = m/2; i >= 1; i--)
    heapfy_mstindex(pq, &me, i, m);

  Mst = gr_create (n);

  for (i = 0, j = 0, nElem = m; i < n-1 && j < m; j++) {
    int a1 = extract_first_mstindex(pq, &me, nElem--);
    arest aresta = arestas[a1];
    assert (aresta.a<=n && aresta.b<=n);
    if (find (dad, aresta.a, aresta.b, 1)) {
      gr_set_arest(Mst, aresta.a, aresta.b,
		   gr_get_cost(G, aresta.a, aresta.b));
      i++;
    }
  }

  for (i = 1; i <= n; i++) {
      graph_node *h, *no;
      no = gr_get_adj_list(G, i);
      h  = no;
      if (!no) continue;
      for (no = no->prox; no; no = no->prox) {
	if (i < no->v)
	  soma += no->custo;
      }
      gr_return_adj_list(h);
  }

  printf ("custo mst:%d\n", soma);
  return Mst;
}

int remove_no_sdleaves(Graph *G, int s[], int d[], int p)
{
  cond_in_set_type ins;
  int i, v, n = gr_nverts(G), change = 0;
  graph_node *node;

  for (i = 1; i <= n; i++) {
    if (gr_vdegree(G,i) == 1) {
      /* node is a leaf */
      ins.n = p;
      if (!(ins.a = s, cond_is_in_set(i, &ins))
	  && !(ins.a = d, cond_is_in_set(i, &ins))) {
	/* this vertice is not a source or destination, so
	   it must be deleted */
	node = gr_get_adj_list(G, i);
	v  = node->prox->v;
	gr_return_adj_list(node);
	gr_remove_edge(G, i, v);
	change=1;
      }
    }
  }
  return change;
}

int mst_bound(int argc, char **argv)
{
  Graph *Mst;
  ppc_instance *inst;

  if (argc < 2) {
    printf("usage: ateams -mst <instance file>\n");
    exit(1);
  }
  inst = read_instance(argv[1]);
  if (!inst)
    return 1;
  if (!dad) 
    mst_init(inst);
  Mst = kruskal(inst->G);
  while (remove_no_sdleaves(Mst, inst->source, inst->dest, inst->p))
    ;
  printf("custo final:%d", gr_forest_cost(Mst));
  return 0;
}
