/************************************************************************
 * PATHS.C
 * Implements useful algoritms to find paths in a graph.
 *
 * C.A. & Allan - JAN/96
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "ateams.h"


/* global shared memory                     */
raiz *solucoes;
/* number of solutions in shared memory     */
int gNumSolutions;
/* true if we are using the Floyd algorithm */
int gUseFloyd = 0;
/* the shortest path and previous matrix    */
int **A, **Pi;
/* pointers to local vectors                */
static int *s, *dist, *verts, *indice;
static int *grau, *adj,*p, *visited;


void print_path(listvert *l);


/* Alloc memory and init the variables used by this module.
   This should be called before any other function in paths.c */
int InitPaths()
{
  int n = gr_nverts(gInst->G), i;
  NewVecS(dist, int, n+1, e0);
  NewVecS(s, int, n+1, e1);
  NewVecS(verts, int, n+1, e2); 
  NewVecS(indice, int, n+1, e3);
  NewVecS(grau, int, n+1, e4);
  NewVecS(p, int, n+1, e5);
  NewVecS(adj, int, n+1, e6);
  NewVecS(A, int*, n+1, e7);
  NewVecS(visited, int, n+1, e8);
  NewVecS(Pi, int*, n+1, e9);
  
  for (i=1; i<=n; i++){
    NewVecS(A[i], int, n+1, e10);
    NewVecS(Pi[i], int, n+1, e11);
  }
  return 1;
  
 e11:
  Delete(A[i]);
 e10:
  for (i--; i>0; i--) {
    Delete(Pi[i]);
    Delete(A[i]);
  }
 e9: Delete(visited);
 e8: Delete(A);
 e7: Delete(adj);
 e6: Delete(p);
 e5: Delete(grau);
 e4: Delete(indice);
 e3: Delete(verts);
 e2: Delete(s);
 e1: Delete(dist);
 e0: fatal_error("error: no memory to paths.c module\n");
  return 0;
}

/* Deallocates memory used by this module */
void  DeletePaths()
{
  int i;
  for (i=gr_nverts(gInst->G); i>0; i--) {
    Delete(Pi[i]);
    Delete(A[i]);
  }
  Delete(visited);
  Delete(A);
  Delete(adj);
  Delete(p);
  Delete(grau);
  Delete(indice);
  Delete(verts);
  Delete(s);
  Delete(dist);
}


typedef struct {
  int *ind;
  int *c;
} sp_heap;


#define shortp int
#define S ((sp_heap *)s)
#define heap_compare(p,q) (S->c[p] < S->c[q])
#define heap_swap(p,q) \
  do { shortp x = a[p]; a[p] = a[q]; a[q] = x; \
    x = S->c[p]; S->c[p] = S->c[q]; S->c[q] = x; \
    S->ind[a[p]]= p; S->ind[a[q]]= q; } while (0)
define_heap_operation(shortp)
#undef heap_compare
#undef heap_swap
#undef shortp
#undef S


/* Uses the Floyd algorithm to find all shortest paths.
   The path is stored implicitly in the matrix Pi */
void all_shortest_paths(Graph *G)
{
  int i, j, k;
  int n = gr_nverts(G);
  graph_node *l, *adj;

  gUseFloyd = 1;
  /* copy values for edges */
  for (i=1; i<=n; i++){
    for (j=1; j<=n; j++)
      A[i][j] = INF;
    memset(Pi[i], 0, sizeof(int)*(n+1));
    adj = gr_get_adj_list(G, i);
    for (l = adj->prox; l; l = l->prox)
      A[i][l->v] = gr_get_cost(G, i, l->v);
    gr_return_adj_list(adj);
  }
  for (k=1; k<=n; k++)
    for (i=1; i<=n; i++)
      for (j=1; j<=n; j++)
	/* there is a difficulty here: INF + INF < 0. Even if I reduce
	   the value of INF the can happen in some place. So, I never
	   do this */
	if (A[i][j] > A[i][k]+A[k][j] && A[i][k]!=INF
	    && A[k][j]!=INF){  
	  A[i][j] = A[i][k]+A[k][j];
	  Pi[i][j] = k;
	}
  for (i=1; i<=n; i++)
    Pi[i][i] = A[i][i] = 0;
}


/* Finds the shortest path between s and d.
   Uses the matrix Pi, caculated by the all_shortest_paths
   function with the Floyd algorithm.
   This turns the complexity to only O(n) */
listvert *shortest_path_floyd(Graph * G, vert source, vert d)
{
  int i=source, j=d, n=0, k;
  listvert *header = NULL, *aux;
  header    = new_list_vert();
  header->v = source;
  aux = header; 
  for (;;){
    while (Pi[i][j] != 0) {
      k = Pi[i][j];
      s[n++]=j;
      j = k;
    }
    if (n==0){
      /* empty queue */
      aux->prox = new_list_vert();
      aux = aux->prox;
      aux->v = d;
      aux->prox = NULL;
      return header;
    }
    i = j;
    j = s[--n];
    aux->prox = new_list_vert();
    aux = aux->prox;
    aux->v = i;
  }
}

listvert *shortest_path(Graph *G, vert v, vert dest)
{
#if 0  
  if (!gUseFloyd)
    return shortest_path_dijkstra(G, v, dest);
  else
#endif    
    return shortest_path_floyd(G, v, dest);
}



/* find the shortest path from one vertex to all others
   unsing Dijkstra's algorithm */
void all_shortest_paths_from_vertex(Graph *G, vert v,
				    int *distances, int *pi)
{
  int i, u, custo, hsize, nVert=gr_nverts(G);
  graph_node *n, *l;
  sp_heap sph;

  memset(pi, 0, sizeof(int)*(nVert+1));
  memset(s,  0, sizeof(int)*(nVert+1));
  sph.c = dist;
  sph.ind = indice;
  for (i = 1; i <= nVert; i++) {
    dist[i]   = INF;
    indice[i] = verts[i] = i;
  }

  dist[1] = 0;
  verts[1] = v;
  indice[1] = v;
  verts[v] = 1;
  indice[v] = 1;
  hsize = nVert;
  for (i=0; i < nVert; i++) {
    /*u = extract_min(verts, dist, indice, hsize--);*/
    u = extract_first_shortp(verts, &sph, hsize--);
    s[u] = 1;
    /* relax the constraints */
    l = gr_get_adj_list(G, u);
    for (n = l->prox; n; n=n->prox) {
      if (s[n->v]==0 && dist[indice[u]] < INF){
	custo = dist[indice[u]] + gr_get_cost(G, u, n->v);
	if (dist[indice[n->v]] > custo) {
	  dist[indice[n->v]] = custo;
	  pi[n->v] = u; 
	  heap_adjust_shortp(verts, &sph, indice[n->v], hsize);
	}
      }
    }
    gr_return_adj_list(l);
  }
  for (i=1; i<=nVert; i++)
    distances[i] = dist[indice[i]];
}


int shortest_path_dist(Graph * G, vert source, vert d)
{
  return A[source][d];
}

listvert *create_path_from_stack(int *p, int v, int top)
{
  listvert *h, *t;
  h = new_list_vert();
  h->v = v;
  h->prox = NULL;
  while (top > -1) {
    t = new_list_vert();
    t->v = p[top--];
    t->prox = h;
    h = t;
  }
  return h;
}

int get_nvert_not_visited(graph_node *node, int *vis)
{
  int n;
  graph_node *t;
  t = node->prox;
  for (n = 0; t; t = t->prox) {
    if (vis[t->v] == 0) {
      n++;
    }
  }
  return n;
}

vert get_vertex_n_not_visited(graph_node *node, int *vis, int pos)
{
  int i, u = 0;
  graph_node *t;
  t=node->prox;
  for (i=0; i<pos; t = t->prox) {
    if (visited[t->v] == 0) {
      i++;
      u = t->v;
    }
  }
  return u;
}




/* Funcao que calcula um caminho qualquer (usando um criterio
 * randomico) entre dois vertices especificados
 * retorna uma lista para o caminho */
#if 0
{
  int i, rnd, v_found = 0, top= - 1, n = gr_nverts(G), adj_sum;
  graph_node *node;

  for (i = 1; i <= n; i++) {
    grau[i] = gr_vdegree(G, i);
  }
  
  memset(visited, 0, sizeof(int)*(n+1));
  visited[v] = 1;
  while (v_found < n-1) {
    node = gr_get_adj_list(G, v);
    adj_sum = get_nvert_not_visited(node, visited);
    if (adj_sum > 0) {
      p[++top] = v;
      rnd = random_int(adj_sum)+1;
      v = get_vertex_n_not_visited(node, visited, rnd);
      visited[v] = 1;      
      v_found++;
      if (v == dest) {
	gr_return_adj_list(node);
	return create_path_from_stack(p, v, top);
      }
    } else {
      v = p[top--];
    }
    gr_return_adj_list(node);
  }
  return 0;
}
#endif


listvert *random_path_until_condition(Graph * G, vert v,
				      void *arg, int (*condition)())
{
  int i, rnd, v_found = 0, top= - 1, n = gr_nverts(G), adj_sum;
  graph_node *node;

  for (i = 1; i <= n; i++) {
    grau[i] = gr_vdegree(G, i);
  }
  
  memset(visited, 0, sizeof(int)*(n+1));
  visited[v] = 1;
  while (v_found < n-1) {
    node = gr_get_adj_list(G, v);
    adj_sum = get_nvert_not_visited(node, visited);
    if (adj_sum > 0) {
      p[++top] = v;
      rnd = random_int(adj_sum)+1;
      v = get_vertex_n_not_visited(node, visited, rnd);
      visited[v] = 1;      
      v_found++;
      if (condition(v, arg)) {
	gr_return_adj_list(node);
	return create_path_from_stack(p, v, top);
      }
    } else {
      if (top<0)
	printf("oi");
      v = p[top--];
    }
    gr_return_adj_list(node);
  }
  return 0;
}

int cond_is_dest(int v, int *dest)
{
  return v == *dest;
}

listvert *random_path(Graph * G, vert v, vert dest)
{
  return random_path_until_condition(G, v, &dest, cond_is_dest);
}

/* ramdom_path Nearest Source ( usa a fonte mais proxima disponvel */


int cond_is_in_set(int v, cond_in_set_type *t)
{
  int i, n = t->n, *a = t->a;
  for (i=0; i<n; i++) {
    if (v == a[i]) {
      return 1;
    }
  }
  return 0;
}

listvert *find_last_source_in_path(listvert *h, int *source, int n_sources)
{
  cond_in_set_type is;
  listvert *t, *p;
  is.n = n_sources;
  is.a = source;
  for (t = h; t; t = t->prox) {
    if (cond_is_in_set(t->v, &is)) {
      p = t;
    }
  }
  return p;
}

void verify_sdfree_path(listvert *h, int *source, int *dest, int p)
{
  cond_in_set_type is;
  listvert *t = h;
  assert(h);
  is.n = p;
  is.a = source;
  for (h = h->prox; h; h = h->prox) {
    assert (!cond_is_in_set(h->v, &is));
  }
  is.a = dest;
  for (h = t; h->prox; h = h->prox) {
    assert (!cond_is_in_set(h->v, &is));
  }  
}

listvert *delete_path_until_pos(listvert *h, listvert *pos)
{
  listvert *t;
  while (h != pos) {
    t = h->prox;
    delete_list_vert(h);
    h = t;
  }
  return h;
}

void delete_path(listvert *h)
{
  delete_path_until_pos(h, 0);
}

listvert *random_path_sdfree(Graph *G, vert *v, vert *select_dest,
			     vert *source, vert *dest, int n_sources)
{
  cond_in_set_type is;
  listvert *h, *p = 0;

  is.n = n_sources;
  is.a = dest;
  h = random_path_until_condition(G, *v, &is, cond_is_in_set);
  p = find_last_source_in_path(h, source, n_sources);
  h = delete_path_until_pos(h, p);
#ifndef NDEBUG
  verify_sdfree_path(h, source, dest, n_sources);
#endif  
  *v = h->v;
  while (p->prox) p = p->prox;
  *select_dest = p->v;  
  return h;
}

void print_path(listvert *l)
{
  while (l){
    printf("%d, ",l->v);
    l = l->prox;
  }
  printf("\n");
}
