#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include "ateams.h"
#include "rf.h"


extern float gLB;

/*********
 * Union-Find implementation for restricted forest
 */

/*   dad is the set of integers stored
 *   returns true if a and b are in the same set
 *   if doit is true, b is put in the same set as a.
 */
int rf_find (rf_find_type *dad, int a, int b, int doit)
{
  int i = a, j = b;
  while (dad->a[i].x != i)
    i = dad->a[i].x;
  while (dad->a[j].x != j)
    j = dad->a[j].x;
  if (doit && (i != j)) {
    dad->a[i].diff += dad->a[j].diff;
    dad->a[i].pt += dad->a[j].pt;
    dad->a[i].nsource += dad->a[j].nsource;
    dad->a[j].x = i;
  }
  return (i == j);
}
/* returns the component in which a is located */
int rf_uf_component(rf_find_type *dad, int a)
{
  int i = a;
  while (dad->a[i].x != i)
    i = dad->a[i].x;
  return i;
}

/* returns 1 if there is an different number of source and
   destination vertices in the component in which v is located.
   Else, returns 0 */
int rf_f(rf_find_type *dad, int a)
{
  int i = a;
  while (dad->a[i].x != i)
    i = dad->a[i].x;
  return dad->a[i].diff != 0;
}

int rf_g(rf_find_type *dad, int a)
{
  int i = a;
  while (dad->a[i].x != i)
    i = dad->a[i].x;
  return dad->a[i].pt != 0;
}

int rf_h(rf_find_type *dad, int a)
{
  int i = a;
  while (dad->a[i].x != i)
    i = dad->a[i].x;
  return ((dad->a[i].diff != 0)
	  || ((dad->a[i].nsource == dad->p) && (dad->a[i].pt != dad->npts))
	  || ((dad->a[i].nsource == 0)      &&  (dad->a[i].pt != 0)));
}


/*
 * binary heap implementation for restricted forest
 */
typedef struct {
  rf_arest *ar;
  int *pos;    /* pos of arest i in heap */
} rfhelp;

#define rfarest int
#define S ((restricted_forest_tad *)s)
#define heap_compare(p,q) (S->ar[a[p]].key < S->ar[a[q]].key)
#define heap_swap(p,q) \
  { int  x = a[p]; a[p] = a[q]; a[q] = x; \
    S->pos[a[p]]= p; S->pos[a[q]]= q; }
define_heap_operation(rfarest)
#undef heap_compare
#undef heap_swap
#undef rfarest
#undef S

/* algoritmo floresta-restrita: */

/*
  Entrada: Grafo G(V,E), custos c_e >= 0, funcao f.
  Sa'ida: Floresta F' e limite inferior LB

  Estruturas de dados:

    C - Conjunto representado pelo TAD uniao-busca
    PQ - Fila de prioridade implementada com um heap bin'ario
    F - Conjunto de arestas representado por um grafo.

  Algoritmo:
    inicialize o grafo F
    LB = 0
    
*/

restricted_forest_tad *
rf_create(ppc_instance *inst, int (*f1)()) 
{
  int n = gr_nverts(inst->G), m = gr_narests(inst->G);
  restricted_forest_tad *rf;

  NewS(rf, restricted_forest_tad, error0);
  NewVecS(rf->C, rf_find_type, 1, error1a);  
  NewVecS(rf->C->a, rf_find_array_type, n+1, error1);
  NewVecS(rf->ar, rf_arest, m+1, error2);
  NewVecS(rf->comp, int, n+1, error3);
  NewVecS(rf->comp_vert, int, n+1, error4);
  NewVecS(rf->index, int, m+1, error5);
  NewVecS(rf->pos, int, m+1, error6);
  NewVecS(rf->d, float, n+1, error7);
  NewVecS(rf->td, int, inst->p, error8);
  NewVecS(rf->part_verts, int, n+1, error9);

  rf->inst = inst;
  rf->f = f1;
  rf->g = f1;
  return rf;

 error9:  Delete(rf->td);
 error8:  Delete(rf->d);
 error7:  Delete(rf->pos);
 error6:  Delete(rf->index);
 error5:  Delete(rf->comp_vert);
 error4:  Delete(rf->comp);
 error3:  Delete(rf->ar);
 error2:  Delete(rf->C->a);
 error1:  Delete(rf->C);
 error1a: Delete(rf);
 error0:  return 0;
}

void rf_delete(restricted_forest_tad *rf)
{
  Delete(rf->part_verts);
  Delete(rf->td);
  Delete(rf->d);
  Delete(rf->pos);
  Delete(rf->index);
  Delete(rf->comp_vert);
  Delete(rf->comp);
  Delete(rf->ar);
  Delete(rf->C->a);
  Delete(rf->C);
  Delete(rf);
}


/* find different arests in Graph G.
   returns the number of different arests found */
int rf_find_arests_in_graph(Graph *G, rf_arest *arestas)
{
  graph_node *h, *no;
  int i, j, n;

  assert(G);
  n = gr_nverts(G);
  
  for (i = 1, j = 1; i <= n; i++) {
    h = no = gr_get_adj_list(G, i);
    if (!no) continue;

    for (no = no->prox; no && no->v < i; no=no->prox) {
      arestas[j].a = i;
      arestas[j].b = no->v;
      arestas[j++].key = no->custo;
    }
    gr_return_adj_list(h);
  }
  return j-1;
}

#define valid_vertex(v, n) ((v)>0 && (v)<=(n))

int get_verts_from_arests(arest *a, int *verts, int n, int m)
{
  int i, total=0;
  memset(verts, 0, sizeof(int)*(n+1));
  for (i=0; i<m; i++) {
    assert (valid_vertex(a[i].a, n) && valid_vertex(a[i].b, n));
    if (verts[a[i].a] == 0) {
      verts[a[i].a] = 1;
      total++;
    }
    if (verts[a[i].b] == 0) {
      verts[a[i].b] = 1;
      total++;
    }
  }
  return total;
}


#define ar_equal(a,b) \
  (((a).a = (b).a) || ((a).b = (b).b))

#define rf_heap_swap(p,q) \
  { int  x = ind[(p)]; ind[(p)] = ind[(q)]; ind[(q)] = x; \
    pos[ind[(p)]]= (p); pos[ind[(q)]]= (q); }


void rf_init_alg1(restricted_forest_tad *rf)
{
  int n = gr_nverts(rf->inst->G);
  int m = gr_narests(rf->inst->G);
  int i, j, ac1, ac2;

  rf->nac = 2*rf->inst->p;
  rf->C->p = rf->inst->p;
  /* put each vertex in it's own component */
  for (i=0; i<=n; i++) {
    rf->C->a[i].x = i;
    rf->C->a[i].diff = 0;
    rf->C->a[i].nsource = 0;
    rf->comp[i] = 0;
    rf->d[i] = 0;
  }
  /* init values for diff */
  for (i=0; i<rf->inst->p; i++) {
    rf->C->a[rf->inst->source[i]].diff = 1;
    rf->C->a[rf->inst->source[i]].nsource = 1;
    rf->C->a[rf->inst->dest[i]].diff = -1;
  }
  for (i=0; i<=m; i++) {
    rf->index[i] = rf->pos[i] = i;
  }
  j = rf_find_arests_in_graph(rf->inst->G, rf->ar);
  assert (j == m);
  /* adjust the initial value for key */
  for (i = 1; i<=m; i++) {
    ac1 = rf->f(rf->C, rf->ar[i].a);
    ac2 = rf->f(rf->C, rf->ar[i].b);
    if ((ac1 + ac2) == 0) {
      rf->ar[i].key = (float)INT_MAX;
    } else {
      rf->ar[i].key /= (float)(ac1 + ac2);
    }
  }
  /* Constructs the priority queue */
  for (i = m/2; i >= 1; i--) {
    heapfy_rfarest(rf->index, rf, i, m);
  }
}

void rf_init_alg(restricted_forest_tad *rf)
{
  int i, m;
  m = gr_narests(rf->inst->G);

  rf_init_alg1(rf);
  /* init values for keys */
  for (i=1; i<=m; i++) {
    rf->ar[i].key = gr_get_cost(rf->inst->G, rf->ar[i].a, rf->ar[i].b);
  }
}


void rf_adjust_key(restricted_forest_tad *rf, int v1, int v2, int i, int m1)
{
  int cost = gr_get_cost(rf->inst->G, v1, v2);
  rf_arest *ar = rf->ar;
  int ac1, ac2;
  
  ac1 = rf->f(rf->C, v1);
  ac2 = rf->f(rf->C, v2);
  if (ac1 == 0 && ac2 == 0) {
    ar[i].key = (float)INT_MAX;
  } else if (ac1 == 1 && ac2 == 1) {
    ar[i].key = cost/2.0;
  } else {
    ar[i].key = cost;
  }
  heap_adjust_rfarest(rf->index, rf, rf->pos[i], m1);
}

void rf_readjust_arests(restricted_forest_tad *rf, int m, int *m1)
{
  int i, a, b;
  for (i=1; i<=m; i++) {
    a = rf->ar[i].a;
    b = rf->ar[i].b;
    /* rf_adjust_key(rf, a, b, i, m1); */
  }
}

/* algoritmo:

   inicialize as variaveis;
   Enquanto houver componentes ativos:
     pegue a aresta que minimize epsilon (que esta' na frente da fila)
     reajuste as arestas da fila
     Adicione a aresta ao grafo
     Atualize d(v)
     Atualize LB
     Atualize o conjunto C
   No final, apage arestas indesejaveis.
*/

Graph *restricted_forest_optimized(restricted_forest_tad *rf, float* LB)
{
  Graph *F, *G = rf->inst->G;
  rf_arest *ar;
  int *ind, *pos, ac1, ac2;
  float *d, epsilon=0, lb=0;
  int a, c, cost, i, i1, i2, j;
  int m, m1, n, v, p, v1, v2;

  /* inicie variaveis */
  assert(G != NULL && rf != NULL);
  n = gr_nverts(G);
  m = gr_narests(G);
  F = gr_create(n);
  m1 = m;
  d   = rf->d; 
  ind = rf->index;
  pos = rf->pos;
  ar  = rf->ar;

  rf_init_alg1(rf);

  /* extract the arest with lower epsilon */
  a = extract_first_rfarest(ind, rf, m1--); 
  while (ar[a].key < (float)INT_MAX){
    v1 = ar[a].a;
    v2 = ar[a].b;
    cost = gr_get_cost(G, v1, v2);
    /* add arest a to F */
    gr_set_arest(F, v1, v2, cost);
    /* computes epsilon */
    ac1 = rf->f(rf->C, v1);
    ac2 = rf->f(rf->C, v2);
    epsilon = (cost - (d[v1] + d[v2]))/(ac1+ac2);

    P(v1)P(v2); /*P(rf->f(rf->C,79))P(rf->f(rf->C,72));*/
    P(cost)P(ac1+ac2)DF(d[v1])DF(d[v2])DF(epsilon)DF(lb)DS("\n");

    /* put the vertices a.a and a.b in the same component */
    rf_find(rf->C, v1, v2, 1);

    /* recompute d(v) */
    for (i=1; i<=n; i++) {
      d[i] += epsilon*(rf->f(rf->C, i));
    }

    /* recompute the number of active components */
    if (rf->f(rf->C, v1) == 0) {
      rf->nac -= ac1 + ac2;
    } else {
      rf->nac -= ac1 + ac2 -1;
    }
    /* computes lower bound */
    lb += epsilon * rf->nac;

#if 0
    rf_readjust_arests(rf, m, m1);
#else    

    for (i=0; i<=n; i++) {
      rf->comp[i] = 0;
    }
    for (i=1; i<=m; i++) {
      if (pos[i] > m1) continue; 

      /* verify if arest i is adjacent to the new component */
      v = 0;
      if (rf_find(rf->C, ar[i].a, v1, 0)) {
	v = ar[i].b;
      } else if (rf_find(rf->C, ar[i].b, v1, 0)){
	v = ar[i].a;
      }
      assert(!rf_find(rf->C, v, v1, 0));
      
      /* if arest i isn't adjacent to a, do nothing */
      if (v == 0) continue;

      c = rf_uf_component(rf->C, v);
      assert(rf->comp[c] != -1);
      if (rf->comp[c] == 0) { 
	/* if there wasn't an arest to this component */
	rf->comp[c] = i;
	rf->comp_vert[c] = v;
      } else { 
	/* this is the second arest */
	i1 = rf->comp[c];
	i2 = i;
	rf->comp[c] = -1;
	if ( ar[i1].key < ar[i2].key ) {
	  swap(int, i1, i2);
	  v = rf->comp_vert[c];
	}
	/* retire a maior aresta da fila */
	p = pos[i1];
	rf_heap_swap(p, m1);
	m1--;
	heap_adjust_rfarest(ind, rf, p, m1);
#if 0
	rf_adjust_key(rf, v1, v, i2, m1);
#else	
	/* ajusta a menor aresta apenas se o componente
	   atual estiver inativo */
	if (rf->f(rf->C, v1) == 0) {
	  if (rf->f(rf->C, v) == 0) {
	    ar[i2].key = (float)INT_MAX;
	  } else {
	    ar[i2].key = ar[i2].key * 2;
	  }
	}
	heap_adjust_rfarest(ind, rf, pos[i2], m1);
#endif	
      }
    }
#if 0    
    for (j=1; j<=n; j++) {
      if (rf->comp[j] > 0) {
	i = rf->comp[j];
	v = rf->comp_vert[j];
	rf_adjust_key(rf, v1, v, i, m1);
      }
    }
#else    
    /* se o componente atual estiver inativo */
    if (rf->f(rf->C, v1) == 0) {
      /* for each arest j marked but not processed */
      for (j=1; j<=n; j++) {
	if (rf->comp[j] > 0) {
	  i = rf->comp[j];
	  v = rf->comp_vert[j];
	  if (rf->f(rf->C, v) == 0) {
	    ar[i].key = (float)INT_MAX;
	  } else {
	    ar[i].key = ar[i].key * 2;
	  }
	  heap_adjust_rfarest(ind, rf, pos[i], m1);
	}
      }
    }
#endif    
#endif    
    /* extract the arest with lower epsilon */
    a = extract_first_rfarest(ind, rf, m1--); 
  }
  *LB = (int)lb;
  return F;
}


int find_best_arest1(restricted_forest_tad *rf,
		     int *v1, int *v2, int *val, float *epsilon)
{
  int a, b, ac1, ac2, i, m , minp=0;
  float cost, ep, min = (float) INT_MAX;
  m = gr_narests(rf->inst->G);
  for (i=1; i<=m; i++){
    a = rf->ar[i].a;
    b = rf->ar[i].b;
    ac1 = rf->f(rf->C, a);
    ac2 = rf->f(rf->C, b);
    /* calc epsilon */
    cost = rf->ar[i].key;
    if (ac1+ac2 != 0) {
      ep = (cost - (rf->d[a] + rf->d[b]))/(ac1+ac2);
    } else {
      ep = (float) INT_MAX - 1;
    }
    if (ep < min
	&& rf_uf_component(rf->C, a) != rf_uf_component(rf->C, b)) {
      min = ep;
      minp = i;
    }
  }
  *epsilon = min;
  *val = rf->ar[minp].key;
  rf->ar[minp].key = (float) INT_MAX;
  *v1 = rf->ar[minp].a;
  *v2 = rf->ar[minp].b;
  return minp;
}

void  rf_remove_no_needed_edges(restricted_forest_tad *rf, Graph *H)
{
#if 0
  int i, n=gr_nverts(H), f;
  printf("removendo\n");
  for (i=1; i<=n; i++) {
    printf("(%d %d %d)", i, rf->C[i].x, rf->C[i].diff);
  } printf("\n\n");
  for (i=1; i<=n; i++) {
    f = i;
    while (rf->C[f].x != f && rf->C[f].diff == 0) {
      if (gr_remove_edge(H, f, rf->C[f].x))
	printf("(%d %d)", f, rf->C[f].x); fflush(stdout);
      f = rf->C[f].x;
    }
  } printf("terminei"); fflush(stdout);
#else 
    while (remove_no_sdleaves(H, rf->inst->source,
			      rf->inst->dest, rf->inst->p))
      ;
#endif
}

void verify_solution(restricted_forest_tad *rf)
{
  int i, n = gr_nverts(rf->inst->G);
  for (i=1; i<=n; i++) {
    if (rf->C->a[i].x == i) {
      assert (rf->C->a[i].diff == 0);
    }
  }
}

Graph *restricted_forest(restricted_forest_tad *rf, float* lb)
{
  Graph *F, *G = rf->inst->G;
  int ac1, ac2, cost, i, n, v1, v2;
  float epsilon=0;

  /* inicie variaveis */
  assert(G != NULL && rf != NULL);

  n = gr_nverts(G);
  F = gr_create(n);
  rf_init_alg(rf);

  while (rf->nac > 0) {
    find_best_arest1(rf, &v1, &v2, &cost, &epsilon);
    gr_set_arest(F, v1, v2, cost);

    /* recompute d(v) */
    for (i=1; i<=n; i++) {
      if (rf->g(rf->C, i) == 1) {
	rf->d[i] += epsilon;
      }
    }

    /* compute lower bound */
    *lb += epsilon * rf->nac;

    ac1 = rf->f(rf->C, v1);
    ac2 = rf->f(rf->C, v2);
    rf->nac -= ac1 + ac2;

    /* put the vertices v1 and v2 in the same component */
    rf_find(rf->C, v1, v2, 1);
    
    /* recompute the number of active components */
    if (rf->f(rf->C, v1) == 1) {
      rf->nac++;
    }
  }
#ifndef NDEBUG  
  verify_solution(rf);
#endif  
  rf_remove_no_needed_edges(rf, F);
  gLB = *lb;
  return F;
}

Graph *restricted_forest_withpt(restricted_forest_tad *rf,
				int *pt, int npts)
{
  int i, n;
  float lb;
  n = gr_nverts(rf->inst->G);
  rf->C->npts = npts;
  for (i=0; i<n; i++) {
    rf->C->a[i].pt = pt[i];
  }
  rf->g = rf_h;
  return restricted_forest(rf, &lb);
}


void rf_exec(restricted_forest_tad *rf, int* LB)
{
  float lb;
  Graph *G;
  G = restricted_forest(rf, &lb);
  printf("lower bound = %f\n", lb);
  printf("solution cost = %d\n", gr_forest_cost(G));
}

int rf_bound(int argc, char **argv)
{
  restricted_forest_tad *rf;
  int lb;
  ppc_instance *inst;

  if (argc < 2) {
    printf("usage: ateams -rf <instance file>\n");
    exit(1);
  }
  inst = read_instance(argv[1]);
  if (!inst)
    return 1;
  rf = rf_create(inst, rf_f);
  rf_exec(rf, &lb);
  return 0;
}

int get_dest_in_same_component(restricted_forest_tad *rf,
			       vert *d, vert source)
{
  vert v;
  int j, c1, c2, p = rf->inst->p;
  c1 = rf_uf_component(rf->C, source);
  for (j=0; j<p; j++) {
    if (d[j] != 0) {
      c2 = rf_uf_component(rf->C, d[j]);
      if (c1 == c2){
	v = d[j];
	d[j] = 0;
	return v;
      }
    }
  }
  return 0;
}

Solution *rf_create_solution_from_graph(restricted_forest_tad *rf,
					Graph *G)
{
  Solution *sol;
  vert *s = rf->inst->source;
  int i, p = rf->inst->p, dest;

  for (i=0; i<p; i++) {
    rf->td[i] = rf->inst->dest[i];
  }
  NewVecS(sol, Solution, p, e0);
  for (i=0; i<p; i++) {
    dest = get_dest_in_same_component(rf, rf->td, s[i]);
    assert(dest != 0);
    sol[i].dest = dest;
    sol[i].path = random_path(G, s[i], dest);
    if (!sol[i].path) goto e1;
  }  
  return sol;

 e1: for (i--; i>=0; i++) {
   delete_path(sol[i].path);
 }
 e0: return 0;
}
