/************************************************************************
 * SOLUTION.C
 * Implementation of abstract type Solution 
 *
 * C.A. & Allan - JAN/95
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <assert.h>
#include "ateams.h"

struct ARC {
  vert a, b;
};

static struct ARC *arcs;
static  Graph *ng;
static int *s, *d;



/* returns an integer in [1, n], with linear distribution of
   probability */
int linear_rand (int n)
{
  return ((int) ((sqrt (1 + 8 * ((int) (rand () % ((1 + n) * n / 2))))
		  - 1) / 2) + 1);
}

/* define a "triangular" random distribution */
int triangular_rand (int n)
{
  int flag = random_int(2);
  int value = linear_rand (n / 2);
  return (flag % 2 ? value : n - value);
}


/* selects a solution to be deleted */
int get_delete_position()
{
  return linear_rand(gNumSolutions - 1);
}


void DeleteSolution (Solution * s, int p)
{
  int i;
  for (i = 0; i < p; i++){
    listvert *l = s[i].path, *aux;
    while (l){
      aux = l;
      l = l->prox;
      delete_list_vert(aux);
    }
  }
  Delete(s);
}

/* Deallocate the main solution memory */
void DeleteSolutions (int p)
{
  int i;
  for (i = 0; i < gNumSolutions; i++)
    {
      DeleteSolution (solucoes[i].sol, p);
    }

  /* free memory used by this module */
  Delete (solucoes);
  Delete (PartialMem.ar);
  Delete (s);
  Delete (d);
  Delete (arcs);
  DeleteHeurist();
  DeletePaths();
  DeleteListvertModule();
}

/* Initializes solutions vector and performs other
   initializations to this module */
void init_solutions ()
{
  /* alloc the vector for solution memory ... */
  solucoes = NewVec (raiz, gNumSolutions);
  if (!solucoes) {
    fatal_error("I couldn't allocate main solution memory\n");
  }
  /* .. and initilize it */
  memset (solucoes, 0, sizeof(raiz) * gNumSolutions);

  /* alloc space for the partial memory */
  PartialMem.ar = NewVec (arest, gInst->p);

  /* alloc some memory to this module */
  if (!(arcs = NewVec (struct ARC, gr_narests((gInst->G))))) {
    fatal_error ("I couldn't allocate memory for arcs!\n");
  }
  if (!(s = NewVec (int, gInst->p)) || !(d = NewVec (int, gInst->p))) {
    fatal_error ("I couldn't allocate memory for vector s and d!\n");
  }
  ng = gr_create(gr_nverts(gInst->G));
  if (!ng) {
    fatal_error("Error: can't alloc memory for graph in InitSolution\n");
  }
  /* init other modules */
  InitHeurist();
  InitPaths();
  InitListvertModule();
}

/* returns 1 if a and b are equal, otherwise returns 0 */
int IsEqual (Solution * a, Solution * b, int p)
{
  int i;
  /* for each path */
  for (i = 0; i < p; i++){
    listvert *l1, *l2;
    if (a[i].dest != b[i].dest)
      return 0;
    l1 = a[i].path;
    l2 = b[i].path;
    while (l1 && l2){
      if (l1->v != l2->v) return 0;
      l1 = l1->prox;
      l2 = l2->prox;
    }
    if (l1 || l2)  return 0;
  }
  return 1; /* the solutions are equal */
}

/* returns a copy of solution s */
Solution *copy_solution (Solution * s, int p)
{
  Solution *sol;
  int i;

  if (!(sol = NewVec (Solution, p))) {
    fatal_error ("I couldn't allocate memory for copy solutions!\n");
  }
  for (i = 0; i < p; i++){
    sol[i].dest = s[i].dest;
    if (s[i].path){
      listvert *l1, *l2;
      sol[i].path = new_list_vert();
      sol[i].path->v = s[i].path->v;
      l1 = s[i].path->prox, l2 = sol[i].path;
      while (l1){
	l2->prox = new_list_vert();
	l2 = l2->prox;
	l2->v = l1->v;
	l1 = l1->prox;
      }
      l2->prox = NULL;
    }
    else
      sol[i].path = NULL;
  }
  return sol;
}

/* Returns the objective function for this solution */
int eval_solution (Solution *s, Graph *G, int p)
{
  int sum = 0, i;
  int a, b, t;
  listvert *l;

  for (i = 0; i < p; i++) {
    for (l = s[i].path; l && l->prox; l=l->prox){
      a = l->v;
      b = l->prox->v;
      if (a > b) {
	t = a; a = b; b = t;
      }
      gr_set_arest_dir(ng, a, b, gr_get_cost(G, a, b));
    }
  }
  sum = gr_get_total_cost(ng);
  gr_remove_all_edges(ng);
  return sum;
}

/* returns in ar all edges in solution sol */
int arests_in_solution(Solution *sol, arest ar[], int p)
{
  int top = -1, i, a, b, t, n = gr_nverts(gInst->G);
  listvert *l;
  graph_node *node, *adj;
  
  for (i = 0; i < p; i++) {
    for (l = sol[i].path; l && l->prox; l=l->prox){
      a = l->v;
      b = l->prox->v;
      if (a > b) {
	t = a; a = b; b = t;
      }
      gr_set_arest_dir(ng, a, b, 1);
    }  
  }
  for (i = 1; i <= n; i++) {
    adj = gr_get_adj_list(ng, i);
    for (node = adj->prox; node; node=node->prox) {
      ar[++top].a = i;
      ar[top].b = node->v;
    }
    gr_return_adj_list(adj);
  }
  gr_remove_all_edges(ng);
  return top;
}

void print_solution(FILE *f, Solution *sol, int p)
{
  int i;
  for (i = 0; i < p; i++) {
    listvert *l = sol[i].path;
    fprintf (f, "\n");
    while (l){
      fprintf (f, "%d,", l->v);
      l = l->prox;
    }
  }
}


/* apply various tests to see if the the solution is valid */ 
int validate_solution(Solution * sol)
{
  listvert *l;
  int i, j, p = gInst->p;

  /* initialize s and d to zero */
  memset (s, 0, sizeof (int) * gInst->p);
  memset (d, 0, sizeof (int) * gInst->p);

  /* validate the dest field of the solution */
  for (i = 0; i < p; i++)
    for (j = 0; j < p; j++)
      if (sol[i].dest == gInst->dest[j])
	d[j]++;

  for (i = 0; i < p; i++) {
    if (d[i]!= 1) {
      printf("gerada por %s ", heuristic[g_heurist_chosen].h_name); 
      fatal_error("Solution has wrong destination");
    }
  }
  /* the same to each source (this field is more tight due to the
     solution structure) */
  for (i = 0; i < p; i++){
    /* Error: a solution must have at least two vertices */
    if (!sol[i].path || !sol[i].path->prox
	|| sol[i].path->v != gInst->source[i]) {
      fatal_error("Solution has wrong sources");
    }
  }

  /* Try to validate each arest of the solution */
  for (i = 0; i < p; i++)
    for (l = sol[i].path; l->prox; l = l->prox){
      /* the vertex doen's exist */
      if (l->v < 1 || l->v > gr_nverts(gInst->G)) {
	fatal_error("Invalid vertex in solution\n");
      }
      if (gr_get_cost(gInst->G, l->v, l->prox->v)==0){
	printf("Error: solution has invalid arest (%d,%d):\n",
	       l->v, l->prox->v);
	print_solution(stdout, sol, p);
	exit(1);
      }
    }

  return 1;
}

void insert_solution (Solution * sol, int pos, int custo)
{
  int j, k;
  /*  Find a place to insert the solution in order */
  for (j = 0; custo > solucoes[j].custo && j < gNumSolutions; j++);
  /* Delete the choosed solution (pos) */
  DeleteSolution (solucoes[pos].sol, gInst->p);
  /* shifts all the solutions beteween the current position and
     that whose solution will be deleted */
  if (j <= pos)
    {
      for (k = pos; k > j; k--)
			solucoes[k] = solucoes[k - 1];
      /*  Insert the solution */
      solucoes[j].custo = custo;
      solucoes[j].sol = sol;
    }
  else
    {
      for (k = pos; k < j - 1; k++)
	solucoes[k] = solucoes[k + 1];
      /*  Insert the solution */
      solucoes[j - 1].custo = custo;
      solucoes[j - 1].sol = sol;
    }
}

void calc_arest_usage(Graph *G, Graph *ng, Solution *sol, int p)
{
  int i, a, b, c;
  listvert *l;
  for (i=0; i<p; i++){
    for (l=sol[i].path; l->prox; l=l->prox){
      a = l->v;
      b = l->prox->v;
      c = gr_get_cost(ng, a, b);
      gr_set_arest(ng, a, b, c+1);
    }
  }
}

